import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import numpy.ma as ma

from matplotlib.artist import Artist
from matplotlib.dates import date2num
from matplotlib.lines import Line2D

from oceans.ocfis import cart2pol


def stick_plot(time, u, v, **kw):
    """
    Parameters
    ----------
    time: list/arrays of datetime objects
    u, v: list/arrays of 2D vector components.

    Returns
    -------
    q: matplotlib's quiver handle for quiverkey.

    Examples
    --------
    >>> from pandas import date_range
    >>> time = date_range(start='1990-11-01 00:00', end='1991-2-1 00:00')
    >>> u = np.sin(0.1 * time.to_julian_date().values) ** 2 -0.5
    >>> v = np.cos(0.1 * time.to_julian_date().values)
    >>> fig, (ax0, ax1, ax2) = plt.subplots(nrows=3, figsize=(10, 6),
    ...                                     sharex=True)
    >>> q = stick_plot(time, u, v, ax=ax0)
    >>> qk = ax0.quiverkey(q, 0.2, 0.65, 1, "1 m s$^{-1}$",
    ...                    labelpos='N', coordinates='axes')
    >>> l = ax1.plot(time.to_pydatetime(), np.sqrt(u**2 + v**2), label='speed')
    >>> l0 = ax2.plot(time.to_pydatetime(), u, label='u')
    >>> l1 = ax2.plot(time.to_pydatetime(), v, label='v')

    Based on Stephane Raynaud's example from:
    https://www.mail-archive.com/matplotlib-users@lists.sourceforge.net/msg18051.html

    """
    from pandas import DatetimeIndex
    from pandas.plotting import register_matplotlib_converters

    register_matplotlib_converters(True)

    width = kw.pop("width", 0.002)
    headwidth = kw.pop("headwidth", 0)
    headlength = kw.pop("headlength", 0)
    headaxislength = kw.pop("headaxislength", 0)
    angles = kw.pop("angles", "uv")
    ax = kw.pop("ax", None)

    if angles != "uv":
        raise AssertionError(
            "Stickplot angles must be `uv` so that"
            "if *U*==*V* the angle of the arrow on"
            "the plot is 45 degrees CCW from the *x*-axis."
        )

    if isinstance(time, DatetimeIndex):
        time = time.to_pydatetime()

    time, u, v = list(map(np.asanyarray, (time, u, v)))
    if not ax:
        fig, ax = plt.subplots()

    q = ax.quiver(
        date2num(time),
        [[0] * len(time)],
        u,
        v,
        angles="uv",
        width=width,
        headwidth=headwidth,
        headlength=headlength,
        headaxislength=headaxislength,
        **kw
    )

    ax.axes.get_yaxis().set_visible(False)
    ax.xaxis_date()
    return q


def landmask(M, color="0.8"):
    """
    Plot land mask.

    Based on trondkristiansen mpl_util.py.
    """
    # Make a constant colormap, default = grey
    constmap = np.matplotlib.colors.ListedColormap([color])

    jmax, imax = M.shape
    # X and Y give the grid cell boundaries,
    # one more than number of grid cells + 1
    # half integers (grid cell centers are integers)
    X = -0.5 + np.arange(imax + 1)
    Y = -0.5 + np.arange(jmax + 1)

    # Draw the mask by pcolor.
    M = ma.masked_where(M > 0, M)
    plt.pcolor(X, Y, M, shading="flat", cmap=constmap)


def level_colormap(levels, cmap=None):
    """
    Make a colormap based on an increasing sequence of levels.

    Based on trondkristiansen.com mpl_util.py.
    """
    # Start with an existing colormap.
    if not cmap:
        cmap = plt.get_cmap()

    # Spread the colors maximally.
    nlev = len(levels)
    S = np.arange(nlev, dtype="float") / (nlev - 1)
    A = cmap(S)

    # Normalize the levels to interval [0, 1].
    levels = np.array(levels, dtype="float")
    L = (levels - levels[0]) / (levels[-1] - levels[0])

    # Make the color dictionary.
    R = [(L[i], A[i, 0], A[i, 0]) for i in range(nlev)]
    G = [(L[i], A[i, 1], A[i, 1]) for i in range(nlev)]
    B = [(L[i], A[i, 2], A[i, 2]) for i in range(nlev)]
    cdict = {"red": tuple(R), "green": tuple(G), "blue": tuple(B)}

    return matplotlib.colors.LinearSegmentedColormap(
        "%s_levels" % cmap.name, cdict, 256
    )


def get_pointsxy(points):
    """
    Return x, y of the given point object.

    """
    return points.get_xdata(), points.get_ydata()


def compass(u, v, **arrowprops):
    """
    Compass draws a graph that displays the vectors with
    components `u` and `v` as arrows from the origin.

    Examples
    --------
    >>> import numpy as np
    >>> u = [+0, -0.5, -0.50, +0.90]
    >>> v = [+1, +0.5, -0.45, -0.85]
    >>> fig, ax = compass(u, v)

    """

    # Create plot.
    fig, ax = plt.subplots(subplot_kw={"polar": True})

    angles, radii = cart2pol(u, v)

    # Arrows or sticks?
    kw = {"arrowstyle": "->"}
    kw.update(arrowprops)
    [
        ax.annotate("", xy=(angle, radius), xytext=(0, 0), arrowprops=kw)
        for angle, radius in zip(angles, radii)
    ]

    ax.set_ylim(0, np.max(radii))

    return fig, ax


def plot_spectrum(data, fs):
    """
    Plots a Single-Sided Amplitude Spectrum of y(t).

    """
    n = len(data)  # Length of the signal.
    k = np.arange(n)
    T = n / fs
    frq = k / T  # Two sides frequency range.
    N = list(range(n // 2))
    frq = frq[N]  # One side frequency range

    # FFT computing and normalization.
    Y = np.fft.fft(data) / n
    Y = Y[N]

    # Plotting the spectrum.
    plt.semilogx(frq, np.abs(Y), "r")
    plt.xlabel("Freq (Hz)")
    plt.ylabel("|Y(freq)|")
    plt.show()


class EditPoints(object):
    """
    Edit points on a graph with the mouse.  Handles only one set of points.

    Key-bindings:
      't' toggle on and off.  (When on, you can move, delete, or add points.)
      'd' delete the point.
      'i' insert a point.

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> fig, ax = plt.subplots(figsize=(6, 6))
    >>> theta = np.arange(0, 2 * np.pi, 0.1)
    >>> r = 1.5
    >>> xs = r * np.cos(theta)
    >>> ys = r * np.sin(theta)
    >>> points = ax.plot(xs, ys, 'ko')
    >>> p = EditPoints(fig, ax, points[0], verbose=True)
    >>> _ = ax.set_title('Click and drag a point to move it')
    >>> _ = ax.axis([-2, 2, -2, 2])

    Based on https://matplotlib.org/examples/event_handling/poly_editor.html

    """

    epsilon = 5  # Maximum pixel distance to count as a point hit.
    showpoint = True

    def __init__(self, fig, ax, points, verbose=False):
        matplotlib.interactive(True)
        if points is None:
            raise RuntimeError("First add points to a figure or canvas.")
        canvas = fig.canvas
        self.ax = ax
        self.dragged = None
        self.points = points
        self.verbose = verbose
        x, y = get_pointsxy(points)
        self.line = Line2D(
            x,
            y,
            marker="o",
            markerfacecolor="r",
            linestyle="none",
            animated=True,
        )
        self.ax.add_line(self.line)

        self._ind = None  # The active point.

        canvas.mpl_connect("draw_event", self.draw_callback)
        canvas.mpl_connect("button_press_event", self.button_press_callback)
        canvas.mpl_connect("key_press_event", self.key_press_callback)
        canvas.mpl_connect(
            "button_release_event", self.button_release_callback
        )
        canvas.mpl_connect("motion_notify_event", self.motion_notify_callback)
        self.canvas = canvas

    def draw_callback(self, event):
        self.background = self.canvas.copy_from_bbox(self.ax.bbox)
        self.ax.draw_artist(self.line)
        self.canvas.blit(self.ax.bbox)

        if self.verbose:
            print("\nDrawing...")  # noqa

    def points_changed(self, points):
        """
        This method is called whenever the points object is called.

        """
        # Only copy the artist props to the line (except visibility).
        vis = self.line.get_visible()
        Artist.update_from(self.line, points)
        # Don't use the points visibility state.
        self.line.set_visible(vis)

        if self.verbose:
            print("\nPoints modified.")  # noqa

    def get_ind_under_point(self, event):
        """
        Get the index of the point under mouse if within epsilon tolerance.

        """
        # Display coordinates.
        arr = self.ax.transData.transform(self.points.get_xydata())
        x, y = arr[:, 0], arr[:, 1]
        d = np.sqrt((x - event.x) ** 2 + (y - event.y) ** 2)
        indseq = np.nonzero(np.equal(d, np.amin(d)))[0]
        ind = indseq[0]
        if self.verbose:
            print("d[ind] {} epsilon {}".format(d[ind], self.epsilon))  # noqa
        if d[ind] >= self.epsilon:
            ind = None

        if self.verbose:
            print(
                "\nClicked at ({}, {})".format(event.xdata, event.ydata)
            )  # noqa
        return ind

    def button_press_callback(self, event):
        """
        Whenever a mouse button is pressed.

        """
        if not self.showpoint:
            return
        if not event.inaxes:
            return
        if not event.button:
            return
        self._ind = self.get_ind_under_point(event)

        # Get point position.
        x, y = get_pointsxy(self.points)
        self.pick_pos = (x[self._ind], y[self._ind])

        if self.verbose:
            print(
                "\nGot point: ({}), ind: {}".format(self.pick_pos, self._ind)
            )  # noqa

    def button_release_callback(self, event):
        """
        Whenever a mouse button is released.

        """
        if not self.showpoint:
            return
        if not event.button:
            return
        self._ind = None
        if self.verbose:
            print("\nButton released.")  # noqa

    def key_press_callback(self, event):
        """
        Whenever a key is pressed.

        """
        if not event.inaxes:
            return
        if event.key == "t":
            self.showpoint = not self.showpoint
            self.line.set_visible(self.showpoint)
            if not self.showpoint:
                self._ind = None

            if self.verbose:
                print("\nToggle {:d}".format(self.showpoint))  # noqa
            return get_pointsxy(self.points)
        elif event.key == "d":
            x, y = get_pointsxy(self.points)
            ind = self.get_ind_under_point(event)
            if ind is not None:
                if self.verbose:
                    print(
                        "\nDeleted ({}, {}) ind: {}".format(
                            x[ind], y[ind], ind
                        )
                    )  # noqa
                x = np.delete(x, ind)
                y = np.delete(y, ind)
                self.points.set_xdata(x)
                self.points.set_ydata(y)
                self.line.set_data(self.points.get_data())
        elif event.key == "i":
            if self.verbose:
                print("Insert point")  # noqa
            xs = self.points.get_xdata()
            ex, ey = event.xdata, event.ydata
            for i in range(len(xs) - 1):
                self.points.set_xdata(np.r_[self.points.get_xdata(), ex])
                self.points.set_ydata(np.r_[self.points.get_ydata(), ey])
                self.line.set_data(self.points.get_data())
                if self.verbose:
                    print("\nInserting: ({}, {})".format(ex, ey))  # noqa
                break

        self.canvas.draw()

    def motion_notify_callback(self, event):
        """
        On mouse movement.

        """
        if not self.showpoint:
            return
        if not self._ind:
            return
        if not event.inaxes:
            return
        if not event.button:
            return
        x, y = get_pointsxy(self.points)
        dx = event.xdata - self.pick_pos[0]
        dy = event.ydata - self.pick_pos[1]
        x[self._ind] = self.pick_pos[0] + dx
        y[self._ind] = self.pick_pos[1] + dy
        if self.verbose:
            print("\nevent.xdata {}".format(event.xdata))  # noqa
            print("\nevent.ydata {}".format(event.ydata))  # noqa
        self.points.set_xdata(x)
        self.points.set_ydata(y)
        self.line.set_data(list(zip(self.points.get_data())))

        self.canvas.restore_region(self.background)
        self.ax.draw_artist(self.line)
        self.canvas.blit(self.ax.bbox)

        if self.verbose:
            print("\nMoving")  # noqa



import os

import numpy as np


def basename(fname):
    return os.path.splitext(os.path.basename(fname))


class match_args_return(object):
    """
    Function decorator to homogenize input arguments and to make the output
    match the original input with respect to scalar versus array, and masked
    versus ndarray.

    """

    def __init__(self, func):
        self.func = func
        self.__doc__ = func.__doc__
        self.__name__ = func.__name__

    def __call__(self, *args, **kw):
        # Check if is array
        self.array = np.any([hasattr(a, "__iter__") for a in args])

        # Check if is masked
        self.masked = np.any([np.ma.isMaskedArray(a) for a in args])
        newargs = [np.ma.atleast_1d(np.ma.masked_invalid(a)) for a in args]
        newargs = [a.astype(np.float) for a in newargs]
        ret = self.func(*newargs, **kw)
        if not self.masked:  # Return a filled array if not masked.
            ret = np.ma.filled(ret, np.nan)
        if not self.array:  # Return scalar if not array.
            ret = ret[0]
        return ret



from datetime import datetime

import numpy as np
import numpy.ma as ma


earth_radius = 6371.0e3


def h2hms(hours):
    """
    Converts hours to hours, minutes, and seconds.

    Examples
    --------
    >>> h2hms(12.51)
    (12.0, 30.0, 36.0)

    """
    hour = np.floor(hours)
    mins = np.remainder(hours, 1.0) * 60.0
    mn = np.floor(mins)
    secs = np.round(np.remainder(mins, 1.0) * 60.0)
    return hour, mn, secs


def hms2h(h, m=None, s=None):
    """
    Converts hours, minutes, and seconds to hours.

    Examples
    --------
    >>> hms2h(12., 30., 36.)
    12.51
    >>> # Or,
    >>> hms2h(123036)
    12.51

    """
    if not m and not s:
        hms = h
        h = np.floor(hms / 10000)
        ms = hms - h * 10000
        m = np.floor(ms / 100)
        s = ms - m * 100
        hours = h + m / 60 + s / 3600
    else:
        hours = h + (m + s / 60) / 60
    return hours


def ms2hms(millisecs):
    """
    Converts milliseconds to integer hour, minute, seconds.

    Examples
    --------
    >>> ms2hms(1e3 * 60)
    (0.0, 1.0, 0.0)

    """
    sec = np.round(millisecs / 1000)
    hour = np.floor(sec / 3600)
    mn = np.floor(np.remainder(sec, 3600) / 60)
    sec = np.round(np.remainder(sec, 60))
    return hour, mn, sec


def julian(y, m=0, d=0, h=0, mi=0, s=0, noon=False):
    """
    Converts Gregorian calendar dates to Julian dates

    USAGE: [j]=julian(y,m,d,h)

    DESCRIPTION:  Converts Gregorian dates to decimal Julian days using the
                  astronomical conversion, but with time zero starting at
                  midnight instead of noon.  In this convention, Julian day
                  2440000 begins at 0000 hours, May 23, 1968. The decimal
                  Julian day, with double precision, yields an accuracy of
                  decimal days of about 0.1 milliseconds.

    If you want Julian days to start and end at noon set `noon` to True.

    INPUT:
    y : year (e.g., 1979) component
    m : month (1-12) component
    d : day (1-31) component of Gregorian date
    hour : hours (0-23)
    min : minutes (0-59)
    sec : decimal seconds or
    h : decimal hours (assumed 0 if absent)

    OUTPUT:
    j : decimal Julian day number

    last revised 1/3/96 by Rich Signell (rsignell@usgs.gov)

    Examples
    --------
    >>> julian(1968, 5, 23, 0)
    array([2440000.])

    """
    y, m, d, h, mi, s = list(map(np.atleast_1d, (y, m, d, h, mi, s)))
    h = hms2h(h, mi, s)

    mo = m + 9
    yr = y - 1
    i = m > 2
    mo[i] = m[i] - 3
    yr[i] = y[i]
    c = np.floor(yr / 100.0)
    yr = yr - c * 100
    j = (
        np.floor((146097 * c) / 4.0)
        + np.floor((1461 * yr) / 4.0)
        + np.floor((153 * mo + 2) / 5.0)
        + d
        + 1721119
    )

    if noon:
        j = j + (h - 12) / 24
    else:
        j = j + h / 24

    return j


def jdrps2jdmat(jd):
    """
    Convert Signell's Julian days to Matlab's Serial day
    matlab's serial date = 1 at 0000 UTC, 1-Jan-0000.

    Examples
    --------
    >>> jdrps2jdmat(2440000)
    array([718941.])

    """
    return jd - julian(0000, 1, 1, 0, 0, 0) + 1


def jdmat2jdrps(jdmat):
    """
    Convert Matlab's Serial Day to Signell's Julian days
    matlab's serial date = 1 at 0000 UTC, 1-Jan-0000.

    Examples
    --------
    >>> jdmat2jdrps(718941)
    array([2440000.])

    """
    return jdmat + julian(0000, 1, 1, 0, 0, 0) - 1


def gregorian(jd, noon=False):
    """
    Converts decimal Julian days to Gregorian dates using the astronomical
    conversion, but with time zero starting at midnight instead of noon.  In
    this convention, Julian day 2440000 begins at 0000 hours, May 23, 1968.
    The Julian day does not have to be an integer, and with Matlab's double
    precision, the accuracy of decimal days is about 0.1 milliseconds.

    INPUT:   jd  = decimal Julian days

    OUTPUT:  gtime = six column Gregorian time matrix, where each row is:
    [yyyy mo da hr mi sec].
    yyyy = year (e.g., 1979)
    mo = month (1-12)
    da = day (1-31)
    hr = hour (0-23)
    mi = minute (0-59)
    sec = decimal seconds
    example: [1990 12 12 0 0 0] is midnight on Dec 12, 1990.

    Examples
    --------
    >>> gregorian(2440000)
    array([[1968.,    5.,   23.,    0.,    0.,    0.]])

    AUTHOR: Rich Signell  (rsignell@usgs.gov)

    """
    # Add 0.2 milliseconds before Gregorian calculation to prevent
    # roundoff error resulting from math operations on time
    # from occasionally representing midnight as
    # (for example) [1990 11 30 23 59 59.99...] instead of [1990 12 1 0 0 0]);
    # If adding a 0.2 ms to time (each time you go back and forth between
    # Julian and Gregorian) bothers you more than the inconvenient
    # representation of Gregorian time at midnight you can comment this
    # line out...

    jd = np.atleast_1d(jd)
    jd = jd + 2.0e-9

    if noon:
        h = np.remainder(jd, 1) * 24 + 12
        i = h >= 24
        jd[i] = jd[i] + 1
        h[i] = h[i] - 24

    secs = np.remainder(jd, 1) * 24 * 3600

    j = np.floor(jd) - 1721119

    ini = 4 * j - 1
    y = np.floor(ini / 146097)
    j = ini - 146097 * y
    ini = np.floor(j / 4)
    ini = 4 * ini + 3
    j = np.floor(ini / 1461)
    d = np.floor(((ini - 1461 * j) + 4) / 4)
    ini = 5 * d - 3
    m = np.floor(ini / 153)
    d = np.floor(((ini - 153 * m) + 5) / 5)
    y = y * 100 + j
    mo = m - 9
    yr = y + 1
    i = m < 10
    mo[i] = m[i] + 3
    yr[i] = y[i]

    hr, mi, sc = s2hms(secs)
    return np.c_[yr, mo, d, hr, mi, sc]


def s2hms(secs):
    """
    Converts seconds to integer hour,minute,seconds
    Usage: hour, min, sec = s2hms(secs)

    Examples
    --------
    >>> s2hms(3600 + 60 + 1)
    (1.0, 1.0, 1)

    """
    hr = np.floor(secs / 3600)
    mi = np.floor(np.remainder(secs, 3600) / 60)
    sc = np.round(np.remainder(secs, 60))

    return hr, mi, sc


# FIXME: STOPPED HERE
def ss2(jd):
    """
    Return Gregorian start and stop dates of Julian day variable
    Usage:  start, stop = ss2(jd)

    """
    start = gregorian(jd[0])
    stop = gregorian(jd[-1])
    return start, stop


def angled(h):
    """
    ANGLED: Returns the phase angles in degrees of a matrix with complex
            elements.

    Usage:
        deg = angled(h)
        h = complex matrix
        deg = angle in math convention (degrees counterclockwise from "east")

    """
    pd = np.angle(h, deg=True)
    return pd


def ij2ind(a, i, j):
    m, n = a.shape
    return m * i - j + 1  # TODO: Check this +1


def ind2ij(a, ind):
    """
    ind2ij returns i, j indices of array.

    """
    m, n = a.shape
    j = np.ceil(ind / m)
    i = np.remainder(ind, m)
    i[i == 0] = m
    return i, j


def rms(u):
    """
    Compute root mean square for each column of matrix u.

    """
    # TODO: use an axis arg.
    if u.ndim > 1:
        m, n = u.shape
    else:
        m = u.size
    return np.sqrt(np.sum(u ** 2) / m)


def z0toCn(z0, H):
    """
    Convert roughness height z0 to Chezy "C" and Manning's "n" which is a
    function of the water depth

    Inputs:
        z0 = roughness height (meters)
        H = water depth (meters) (can be vector)
    Outputs:
        C = Chezy "C" (non-dimensional)
        n = Manning's "n" (non-dimensional)

    Examples
    --------
    >>> # finds vectors C and n corresponding to a z0=0.003
    >>> # and a range of water depths from 2--200 meters.
    >>> C, n = z0toCn(0.003, np.arange(2, 200))

    """

    k_s = 30 * z0
    C = 18 * np.log10(12 * H / k_s)
    n = (H ** (1.0 / 6.0)) / C

    return C, n


def z0tocd(z0, zr):
    """
    Calculates CD at a given ZR corresponding to Z0.

    """
    cd = (0.4 * np.ones(z0.size) / np.log(zr / z0)) ** 2
    return cd


def short_calc(amin, amax):
    rang = 32767 - (-32767)
    add_offset = (amax + amin) * 0.5
    scale_factor = (amax - amin) / rang
    return add_offset, scale_factor


def gsum(x, **kw):
    """
    Just like sum, except that it skips over bad points.

    """
    xnew = ma.masked_invalid(x)
    return np.sum(xnew, **kw)


def gmean(x, **kw):
    """
    Just like mean, except that it skips over bad points.

    """
    xnew = ma.masked_invalid(x)
    return np.mean(xnew, **kw)


def gmedian(x, **kw):
    """
    Just like median, except that it skips over bad points.

    """
    xnew = ma.masked_invalid(x)
    return np.median(xnew, **kw)


def gmin(x, **kw):
    """
    Just like min, except that it skips over bad points.

    """
    xnew = ma.masked_invalid(x)
    return np.min(xnew, **kw)


def gmax(x, **kw):
    """
    Just like max, except that it skips over bad points.

    """
    xnew = ma.masked_invalid(x)
    return np.max(xnew, **kw)


def gstd(x, **kw):
    """
    Just like std, except that it skips over bad points.

    """
    xnew = ma.masked_invalid(x)
    return np.std(xnew, **kw)


def near(x, x0, n=1):
    """
    Given an 1D array `x` and a scalar `x0`, returns the `n` indices of the
    element of `x` closest to `x0`.

    """
    distance = np.abs(x - x0)
    index = np.argsort(distance)
    return index[:n], distance[index[:n]]


def swantime(a):
    """
    Converts SWAN default time format to datetime object.

    """
    if isinstance(a, str):
        a = float(a)
        a = np.asanyarray(a)

    year = np.floor(a / 1e4)
    a = a - year * 1e4
    mon = np.floor(a / 1e2)
    a = a - mon * 1e2
    day = np.floor(a)
    a = a - day
    hour = np.floor(a * 1e2)
    a = a - hour / 1e2
    mn = np.floor(a * 1e4)
    a = a - mn / 1e4
    sec = np.floor(a * 1e6)

    return datetime(year, mon, day, hour, mn, sec)


def shift(a, b, n):
    """
    a and b are vectors
    n is number of points of a to cut off
    anew and bnew will be the same length.

    """
    # la, lb = a.size, lb = b.size

    anew = a[list(range(0 + n, len(a))), :]

    if len(anew) > len(b):
        anew = anew[list(range(0, len(b))), :]
        bnew = b
    else:
        bnew = b[list(range(0, len(anew))), :]
    return anew, bnew


def lagcor(a, b, n):
    """
    Finds lagged correlations between two series.
    a and b are two column vectors
    n is range of lags
    cor is correlation as fn of lag.

    """
    cor = []
    for k in range(0, n + 1):
        d1, d2 = shift(a, b, k)
        ind = ~np.isnan(d1 + d2)
        c = np.corrcoef(d1[ind], d2[ind])
        if len(c) > 1:
            cor.append(c[0, 1])

    return cor


def coast2bln(coast, bln_file):
    """
    Converts a matlab coast (two column array w/ nan for line breaks) into
    a Surfer blanking file.

    Where coast is a two column vector and bln_file is the output file name.

    Needs `fixcoast`.

    """

    c2 = fixcoast(coast)
    ind = np.where(np.isnan(c2[:, 0]))[0]
    n = len(ind) - 1
    bln = c2.copy()

    for k in range(0, n - 1):
        kk = list(range(ind[k] + 1, ind[k + 1]))
        NP = int(len(kk))
        bln[ind[k], 0] = NP
        bln[ind[k], 1] = int(1)

    bln = bln[:-1]
    np.savetxt(bln_file, bln, fmt="%g")


def fixcoast(coast):
    """
    FIXCOAST  Makes sure coastlines meet Signell's conventions.

    Fixes coastline is in the format we want.  Assumes that lon/lat are in the
    first two columns of the matrix coast, and that coastline segments are
    separated by rows of NaNs (or -99999s).  This routine ensures that only 1
    row of NaNs separates each segment, and makes sure that the first and last
    rows contain NaNs.

    """

    ind = coast == -99999.0
    coast[ind] = np.NaN

    ind = np.where(np.isnan(coast[:, 0]))[0]
    dind = np.diff(ind)
    idup = np.where(dind == 1)[0]

    coast = np.delete(coast, ind[idup], axis=0)

    if not np.isnan(coast[0, 0]):
        coast = np.insert(coast, 0, np.NaN, axis=0)

    if not np.isnan(coast[-1, -1]):
        coast = np.append(coast, np.c_[np.NaN, np.NaN], axis=0)

    return coast



from oceans.sw_extras.gamma_GP_from_SP_pt import gamma_GP_from_SP_pt
from oceans.sw_extras.sw_extras import (
    N,
    cor_beta,
    cph,
    cr_depth,
    inertial_period,
    kdpar,
    o2sol_SP_pt_benson_krause_84,
    photic_depth,
    psu2ppt,
    richnumb,
    shear,
    sigma_t,
    sigmatheta,
    soundspeed,
    spice,
    strat_period,
    tcond,
    visc,
    zmld_boyer,
    zmld_so,
)
from oceans.sw_extras.waves import Waves


__all__ = [
    N,
    cor_beta,
    cph,
    cr_depth,
    inertial_period,
    kdpar,
    o2sol_SP_pt_benson_krause_84,
    photic_depth,
    psu2ppt,
    richnumb,
    shear,
    sigma_t,
    sigmatheta,
    soundspeed,
    spice,
    strat_period,
    tcond,
    visc,
    zmld_boyer,
    zmld_so,
    gamma_GP_from_SP_pt,
    Waves,
]



import os

import numpy as np


_default_path = os.path.join(os.path.dirname(__file__), "data")


def LineNormals2D(Vertices, Lines):
    """
    This function calculates the normals, of the line points using the
    neighboring points of each contour point, and forward an backward
    differences on the end points.

    N = LineNormals2D(V, L)

    inputs,
      V : List of points/vertices 2 x M
    (optional)
      Lines : A N x 2 list of line pieces, by indices of the vertices
            (if not set assume Lines=[1 2; 2 3 ; ... ; M-1 M])

    outputs,
      N : The normals of the Vertices 2 x M

    Examples
    --------
    >>> import numpy as np
    >>> import matplotlib.pyplot as plt
    >>> data = np.load(os.path.join(_default_path, 'testdata.npz'))
    >>> Lines, Vertices = data['Lines'], data['Vertices']
    >>> N = LineNormals2D(Vertices, Lines)
    >>> fig, ax = plt.subplots(nrows=1, ncols=1)
    >>> _ = ax.plot(np.c_[Vertices[:, 0], Vertices[:,0 ] + 10 * N[:, 0]].T,
    ...            np.c_[Vertices[:, 1], Vertices[:, 1] + 10 * N[:, 1]].T)

    Function based on LineNormals2D.m written by
    D.Kroon University of Twente (August 2011)

    """
    eps = np.spacing(1)

    if isinstance(Lines, np.ndarray):
        pass
    elif not Lines:
        Lines = np.c_[
            np.arange(1, Vertices.shape[0]),
            np.arange(2, Vertices.shape[0] + 1),
        ]
    else:
        raise ValueError(f"Expected np.array but got {Lines:!r}.")

    # Calculate tangent vectors.
    DT = Vertices[Lines[:, 0] - 1, :] - Vertices[Lines[:, 1] - 1, :]

    # Make influence of tangent vector 1/Distance (Weighted Central
    # Differences.  Points which are closer give a more accurate estimate of
    # the normal).
    LL = np.sqrt(DT[:, 0] ** 2 + DT[:, 1] ** 2)
    DT[:, 0] = DT[:, 0] / np.maximum(LL ** 2, eps)
    DT[:, 1] = DT[:, 1] / np.maximum(LL ** 2, eps)

    D1 = np.zeros_like(Vertices)
    D2 = np.zeros_like(Vertices)
    D1[Lines[:, 0] - 1, :] = DT
    D2[Lines[:, 1] - 1, :] = DT
    D = D1 + D2

    # Normalize the normal.
    LL = np.sqrt(D[:, 0] ** 2 + D[:, 1] ** 2)
    N = np.zeros_like(D)
    N[:, 0] = -D[:, 1] / LL
    N[:, 1] = D[:, 0] / LL

    return N


def LineCurvature2D(Vertices, Lines=None):
    """
    This function calculates the curvature of a 2D line. It first fits
    polygons to the points. Then calculates the analytical curvature from
    the polygons.

    k = LineCurvature2D(Vertices,Lines)

    inputs,
      Vertices : A M x 2 list of line points.
      (optional)
      Lines : A N x 2 list of line pieces, by indices of the vertices
            (if not set assume Lines=[1 2; 2 3 ; ... ; M-1 M])

    outputs,
      k : M x 1 Curvature values

    Examples
    --------
    >>> import numpy as np
    >>> import matplotlib.pyplot as plt
    >>> data = np.load(os.path.join(_default_path, 'testdata.npz'))
    >>> Lines, Vertices = data['Lines'], data['Vertices']
    >>> k = LineCurvature2D(Vertices, Lines)
    >>> N = LineNormals2D(Vertices, Lines)
    >>> k = k * 100
    >>> fig, ax = plt.subplots(nrows=1, ncols=1)
    >>> _ = ax.plot(np.c_[Vertices[:, 0], Vertices[:, 0] + k * N[:, 0]].T,
    ...             np.c_[Vertices[:, 1], Vertices[:, 1] + k * N[:, 1]].T, 'g')
    >>> _ = ax.plot(np.c_[Vertices[Lines[:, 0] - 1, 0],
    ...                   Vertices[Lines[:, 1] - 1, 0]].T,
    ...             np.c_[Vertices[Lines[:, 0] - 1, 1],
    ...                   Vertices[Lines[:, 1] - 1, 1]].T, 'b')
    >>> _ = ax.plot(Vertices[:, 0], Vertices[:, 1], 'r.')

    Function based on LineCurvature2D.m written by
    D.Kroon University of Twente (August 2011)

    """
    # If no line-indices, assume a x[0] connected with x[1], x[2] with x[3].
    if isinstance(Lines, np.ndarray):
        pass
    elif not Lines:
        Lines = np.c_[
            np.arange(1, Vertices.shape[0]),
            np.arange(2, Vertices.shape[0] + 1),
        ]
    else:
        raise ValueError("Cannot recognized {!r}.".format(Lines))

    # Get left and right neighbor of each points.
    Na = np.zeros(Vertices.shape[0], dtype=np.int)
    Nb = np.zeros_like(Na)
    # As int because we use it to index an array...

    Na[Lines[:, 0] - 1] = Lines[:, 1]
    Nb[Lines[:, 1] - 1] = Lines[:, 0]

    # Check for end of line points, without a left or right neighbor.
    checkNa = Na == 0
    checkNb = Nb == 0

    Naa, Nbb = Na, Nb

    Naa[checkNa] = np.where(checkNa)[0]
    Nbb[checkNb] = np.where(checkNb)[0]

    # If no left neighbor use two right neighbors, and the same for right.
    Na[checkNa] = Nbb[Nbb[checkNa]]
    Nb[checkNb] = Naa[Naa[checkNb]]

    # ... Also, I remove `1` to get python indexing correctly.
    Na -= 1
    Nb -= 1

    # Correct for sampling differences.
    Ta = -np.sqrt(np.sum((Vertices - Vertices[Na, :]) ** 2, axis=1))
    Tb = np.sqrt(np.sum((Vertices - Vertices[Nb, :]) ** 2, axis=1))

    # If no left neighbor use two right neighbors, and the same for right.
    Ta[checkNa] = -Ta[checkNa]
    Tb[checkNb] = -Tb[checkNb]

    x = np.c_[Vertices[Na, 0], Vertices[:, 0], Vertices[Nb, 0]]
    y = np.c_[Vertices[Na, 1], Vertices[:, 1], Vertices[Nb, 1]]
    M = np.c_[
        np.ones_like(Tb),
        -Ta,
        Ta ** 2,
        np.ones_like(Tb),
        np.zeros_like(Tb),
        np.zeros_like(Tb),
        np.ones_like(Tb),
        -Tb,
        Tb ** 2,
    ]

    invM = inverse3(M)
    a = np.zeros_like(x)
    b = np.zeros_like(a)
    a[:, 0] = (
        invM[:, 0, 0] * x[:, 0]
        + invM[:, 1, 0] * x[:, 1]
        + invM[:, 2, 0] * x[:, 2]
    )

    a[:, 1] = (
        invM[:, 0, 1] * x[:, 0]
        + invM[:, 1, 1] * x[:, 1]
        + invM[:, 2, 1] * x[:, 2]
    )

    a[:, 2] = (
        invM[:, 0, 2] * x[:, 0]
        + invM[:, 1, 2] * x[:, 1]
        + invM[:, 2, 2] * x[:, 2]
    )

    b[:, 0] = (
        invM[:, 0, 0] * y[:, 0]
        + invM[:, 1, 0] * y[:, 1]
        + invM[:, 2, 0] * y[:, 2]
    )

    b[:, 1] = (
        invM[:, 0, 1] * y[:, 0]
        + invM[:, 1, 1] * y[:, 1]
        + invM[:, 2, 1] * y[:, 2]
    )

    b[:, 2] = (
        invM[:, 0, 2] * y[:, 0]
        + invM[:, 1, 2] * y[:, 1]
        + invM[:, 2, 2] * y[:, 2]
    )

    # Calculate the curvature from the fitted polygon.
    k = (
        2
        * (a[:, 1] * b[:, 2] - a[:, 2] * b[:, 1])
        / ((a[:, 1] ** 2 + b[:, 1] ** 2) ** (3 / 2))
    )

    return k


def inverse3(M):
    """
    This function does inv(M), but then for an array of 3x3 matrices.

    """
    adjM = np.zeros((M.shape[0], 3, 3))
    adjM[:, 0, 0] = M[:, 4] * M[:, 8] - M[:, 7] * M[:, 5]
    adjM[:, 0, 1] = -(M[:, 3] * M[:, 8] - M[:, 6] * M[:, 5])
    adjM[:, 0, 2] = M[:, 3] * M[:, 7] - M[:, 6] * M[:, 4]
    adjM[:, 1, 0] = -(M[:, 1] * M[:, 8] - M[:, 7] * M[:, 2])
    adjM[:, 1, 1] = M[:, 0] * M[:, 8] - M[:, 6] * M[:, 2]
    adjM[:, 1, 2] = -(M[:, 0] * M[:, 7] - M[:, 6] * M[:, 1])
    adjM[:, 2, 0] = M[:, 1] * M[:, 5] - M[:, 4] * M[:, 2]
    adjM[:, 2, 1] = -(M[:, 0] * M[:, 5] - M[:, 3] * M[:, 2])
    adjM[:, 2, 2] = M[:, 0] * M[:, 4] - M[:, 3] * M[:, 1]

    detM = (
        M[:, 0] * M[:, 4] * M[:, 8]
        - M[:, 0] * M[:, 7] * M[:, 5]
        - M[:, 3] * M[:, 1] * M[:, 8]
        + M[:, 3] * M[:, 7] * M[:, 2]
        + M[:, 6] * M[:, 1] * M[:, 5]
        - M[:, 6] * M[:, 4] * M[:, 2]
    )

    return adjM / detM[:, None, None]



import numpy as np


def scaloa(xc, yc, x, y, t=None, corrlen=None, err=None, zc=None):
    """
    Scalar objective analysis.  Interpolates t(x, y) into tp(xc, yc)
    Assumes spatial correlation function to be isotropic and Gaussian in the
    form of: C = (1 - err) * np.exp(-d**2 / corrlen**2) where:
    d : Radial distance from the observations.

    Parameters
    ----------
    corrlen : float
              Correlation length.
    err     : float
              Random error variance (epsilon in the papers).

    Return
    ------
    tp : array
         Gridded observations.
    ep : array
         Normalized mean error.

    Examples
    --------
    See https://ocefpaf.github.io/python4oceanographers/blog/2014/10/27/OI/

    Notes
    -----
    The funcion `scaloa` assumes that the user knows `err` and `corrlen` or
    that these parameters where chosen arbitrary.  The usual guess are the
    first baroclinic Rossby radius for `corrlen` and 0.1 e 0.2 to the sampling
    error.

    """

    n = len(x)
    x, y = np.reshape(x, (1, n)), np.reshape(y, (1, n))

    # Squared distance matrix between the observations.
    d2 = (np.tile(x, (n, 1)).T - np.tile(x, (n, 1))) ** 2 + (
        np.tile(y, (n, 1)).T - np.tile(y, (n, 1))
    ) ** 2

    nv = len(xc)
    xc, yc = np.reshape(xc, (1, nv)), np.reshape(yc, (1, nv))

    # Squared distance between the observations and the grid points.
    dc2 = (np.tile(xc, (n, 1)).T - np.tile(x, (nv, 1))) ** 2 + (
        np.tile(yc, (n, 1)).T - np.tile(y, (nv, 1))
    ) ** 2

    # Correlation matrix between stations (A) and cross correlation (stations
    # and grid points (C)).
    A = (1 - err) * np.exp(-d2 / corrlen ** 2)
    C = (1 - err) * np.exp(-dc2 / corrlen ** 2)

    if 0:  # NOTE: If the parameter zc is used (`scaloa2.m`)
        A = (1 - d2 / zc ** 2) * np.exp(-d2 / corrlen ** 2)
        C = (1 - dc2 / zc ** 2) * np.exp(-dc2 / corrlen ** 2)

    # Add the diagonal matrix associated with the sampling error.  We use the
    # diagonal because the error is assumed to be random.  This means it just
    # correlates with itself at the same place.
    A = A + err * np.eye(len(A))

    # Gauss-Markov to get the weights that minimize the variance (OI).
    tp = None
    if t:
        t = np.reshape(t, (n, 1))
        tp = np.dot(C, np.linalg.solve(A, t))
        if 0:  # NOTE: `scaloa2.m`
            mD = np.sum(np.linalg.solve(A, t)) / np.sum(
                np.sum(np.linalg.inv(A))
            )
            t = t - mD
            tp = C * (np.linalg.solve(A, t))
            tp = tp + mD * np.ones(tp.shape)
    if not t:
        print("Computing just the interpolation errors.")  # noqa

    # Normalized mean error.  Taking the squared root you can get the
    # interpolation error in percentage.
    ep = 1 - np.sum(C.T * np.linalg.solve(A, C.T), axis=0) / (1 - err)

    return tp, ep



import gsw
import numpy as np


class Waves(object):
    r"""
    Solves the wave dispersion relationship via Newton-Raphson.

    .. math::
        \omega^2 = gk\tanh kh

    Parameters
    ----------
    h : array_like, str
        Water depth [m] or 'deep', 'shallow' as keywords
    T : array_like
        Wave period [s]
    L : array_like
        Wave length [m]
    thetao : array_like
        TODO
    Ho : array_like
        TODO

    Returns
    -------
    omega : array_like
            Wave frequency
    TODO: hoLo, hoL, Lo, L, k, T, Co, C, Cg, G, Ks, Kr, theta, H

    Notes
    -----
    Compare values with:
    http://www.coastal.udel.edu/faculty/rad/wavetheory.html

    Examples
    --------
    >>> from oceans.sw_extras.waves import Waves
    >>> wav = Waves(h=10, T=5, L=None)
    >>> print("ho/Lo = %s" % wav.hoLo)
    ho/Lo = 0.2561951195588007
    >>> print("ho/L  = %s" % wav.hoL)
    ho/L  = 0.2732735643784346
    >>> print("Lo    = %s" % wav.Lo)
    Lo    = 39.03274979328733
    >>> print("L     = %s" % wav.L)
    L     = 36.59336761221369
    >>> print("k     = %s" % wav.k)
    k     = 0.17170284445431747
    >>> print("omega = %s" % wav.omega)
    omega = 1.2566370614359172
    >>> print("T     = %s" % wav.T)
    T     = 5.0
    >>> print("C     = %s" % wav.C)
    C     = 7.318673522442738
    >>> print("Cg    = %s" % wav.Cg)
    Cg    = 4.470858193067349
    >>> print("G     = %s" % wav.G)
    G     = 0.22176735425004176
    >>> wav = Waves(h=10, T=None, L=100)
    >>> print("ho/Lo = %s" % wav.hoLo)
    ho/Lo = 0.05568933069002106
    >>> print("ho/L  = %s" % wav.hoL)
    ho/L  = 0.1
    >>> print("Lo    = %s" % wav.Lo)
    Lo    = 179.56760973950605
    >>> print("L     = %s" % wav.L)
    L     = 100.0
    >>> print("k     = %s" % wav.k)
    k     = 0.06283185307179587
    >>> print("omega = %s" % wav.omega)
    omega = 0.5858823798813203
    >>> print("T     = %s" % wav.T)
    T     = 10.724311778163298
    >>> print("C     = %s" % wav.C)
    C     = 9.324607682855573
    >>> print("Cg    = %s" % wav.Cg)
    Cg    = 8.291208888683515
    >>> print("G     = %s" % wav.G)
    G     = 0.7783501828024171
    >>> print("Ks  = %s" % wav.Ks)
    Ks  = 1.00485953746193

    """

    def __init__(self, h, T=None, L=None, thetao=None, Ho=None, lat=None):
        self.T = np.asarray(T, dtype=np.float)
        self.L = np.asarray(L, dtype=np.float)
        self.Ho = np.asarray(Ho, dtype=np.float)
        self.lat = np.asarray(lat, dtype=np.float)
        self.thetao = np.asarray(thetao, dtype=np.float)

        if isinstance(h, str):
            if L is not None:
                if h == "deep":
                    self.h = self.L / 2.0
                elif h == "shallow":
                    self.h = self.L * 0.05
        else:
            self.h = np.asarray(h, dtype=np.float)

        if lat is None:
            g = 9.81  # Default gravity.
        else:
            g = gsw.grav(lat, p=0)

        if L is None:
            self.omega = 2 * np.pi / self.T
            self.Lo = (g * self.T ** 2) / 2 / np.pi
            # Returns wavenumber of the gravity wave dispersion relation using
            # newtons method. The initial guess is shallow water wavenumber.
            self.k = self.omega / np.sqrt(g)
            # TODO: May change to,
            # self.k = self.w ** 2 / (g * np.sqrt(self.w ** 2 * self.h / g))
            f = g * self.k * np.tanh(self.k * self.h) - self.omega ** 2

            while np.abs(f.max()) > 1e-10:
                dfdk = g * self.k * self.h * (
                    1 / (np.cosh(self.k * self.h))
                ) ** 2 + g * np.tanh(self.k * self.h)
                self.k = self.k - f / dfdk
                # FIXME:
                f = g * self.k * np.tanh(self.k * self.h) - self.omega ** 2

            self.L = 2 * np.pi / self.k
            if isinstance(h, str):
                if h == "deep":
                    self.h = self.L / 2.0
                elif h == "shallow":
                    self.h = self.L * 0.05
        else:
            self.Lo = self.L / np.tanh(2 * np.pi * self.h / self.L)
            self.k = 2 * np.pi / self.L
            self.T = np.sqrt(2 * np.pi * self.Lo / g)
            self.omega = 2 * np.pi / self.T

        self.hoL = self.h / self.L
        self.hoLo = self.h / self.Lo
        self.C = self.omega / self.k  # or L / T
        self.Co = self.Lo / self.T
        self.G = 2 * self.k * self.h / np.sinh(2 * self.k * self.h)
        self.n = (1 + self.G) / 2
        self.Cg = self.n * self.C
        self.Ks = np.sqrt(1 / (1 + self.G) / np.tanh(self.k * self.h))

        if thetao is None:
            self.theta = np.NaN
            self.Kr = np.NaN
        if thetao is not None:
            self.theta = np.rad2deg(
                np.asin(self.C / self.Co * np.sin(np.deg2rad(self.thetao)))
            )
            self.Kr = np.sqrt(
                np.cos(np.deg2rad(self.thetao))
                / np.cos(np.deg2rad(self.theta))
            )

        if Ho is None:
            self.H = np.NaN
        if Ho is not None:
            self.H = self.Ho * self.Ks * self.Kr



from copy import copy

import numpy as np
import seawater as sw

from seawater.constants import OMEGA, earth_radius


def sigma_t(s, t, p):
    """
    :math:`\\sigma_{t}` is the remainder of subtracting 1000 kg m :sup:`-3`
    from the density of a sea water sample at atmospheric pressure.

    Parameters
    ----------
    s(p) : array_like
           salinity [psu (PSS-78)]
    t(p) : array_like
           temperature [:math:`^\\circ` C (ITS-90)]
    p : array_like
        pressure [db]

    Returns
    -------
    sgmt : array_like
           density  [kg m :sup:`3`]

    Notes
    -----
    Density of Sea Water using UNESCO 1983 (EOS 80) polynomial.

    Examples
    --------
    >>> # Data from UNESCO Tech. Paper in Marine Sci. No. 44, p22.
    >>> from seawater.library import T90conv
    >>> import oceans.sw_extras.sw_extras as swe
    >>> s = [0, 0, 0, 0, 35, 35, 35, 35]
    >>> t = T90conv([0, 0, 30, 30, 0, 0, 30, 30])
    >>> p = [0, 10000, 0, 10000, 0, 10000, 0, 10000]
    >>> swe.sigma_t(s, t, p)
    array([-0.157406  , 45.33710972, -4.34886626, 36.03148891, 28.10633141,
           70.95838408, 21.72863949, 60.55058771])

    References
    ----------
    Fofonoff, P. and Millard, R.C. Jr UNESCO 1983. Algorithms for
    computation of fundamental properties of seawater. UNESCO Tech. Pap. in
    Mar. Sci., No. 44, 53 pp.  Eqn.(31) p.39.
    http://www.scor-int.org/Publications.htm

    Millero, F.J., Chen, C.T., Bradshaw, A., and Schleicher, K. A new
    high pressure equation of state for seawater. Deap-Sea Research., 1980,
    Vol27A, pp255-264. doi:10.1016/0198-0149(80)90016-3

    """
    s, t, p = list(map(np.asanyarray, (s, t, p)))
    return sw.dens(s, t, p) - 1000.0


def sigmatheta(s, t, p, pr=0):
    """
    :math:`\\sigma_{\\theta}` is a measure of the density of ocean water
    where the quantity :math:`\\sigma_{t}` is calculated using the potential
    temperature (:math:`\\theta`) rather than the in situ temperature and
    potential density of water mass relative to the specified reference
    pressure.

    Parameters
    ----------
    s(p) : array_like
           salinity [psu (PSS-78)]
    t(p) : array_like
           temperature [:math:`^\\circ` C (ITS-90)]
    p : array_like
        pressure [db]
    pr : number
         reference pressure [db], default = 0

    Returns
    -------
    sgmte : array_like
           density  [kg m :sup:`3`]

    Examples
    --------
    >>> # Data from UNESCO Tech. Paper in Marine Sci. No. 44, p22.
    >>> from seawater.library import T90conv
    >>> import oceans.sw_extras.sw_extras as swe
    >>> s = [0, 0, 0, 0, 35, 35, 35, 35]
    >>> t = T90conv([0, 0, 30, 30, 0, 0, 30, 30])
    >>> p = [0, 10000, 0, 10000, 0, 10000, 0, 10000]
    >>> swe.sigmatheta(s, t, p)
    array([-0.157406  , -0.20476006, -4.34886626, -3.63884068, 28.10633141,
           28.15738545, 21.72863949, 22.59634627])

    References
    ----------
    Fofonoff, P. and Millard, R.C. Jr UNESCO 1983. Algorithms for
    computation of fundamental properties of seawater. UNESCO Tech. Pap. in
    Mar. Sci., No. 44, 53 pp.  Eqn.(31) p.39.
    http://www.scor-int.org/Publications.htm

    Millero, F.J., Chen, C.T., Bradshaw, A., and Schleicher, K. A new
    high pressure equation of state for seawater. Deap-Sea Research., 1980,
    Vol27A, pp255-264. doi:10.1016/0198-0149(80)90016-3

    """
    s, t, p, pr = list(map(np.asanyarray, (s, t, p, pr)))
    return sw.pden(s, t, p, pr) - 1000.0


def N(bvfr2):
    """
    Buoyancy frequency is the frequency with which a parcel or particle of
    fluid displaced a small vertical distance from its equilibrium position in
    a stable environment will oscillate. It will oscillate in simple harmonic
    motion with an angular frequency defined by

    .. math:: N = \\left(\\frac{-g}{\\sigma_{\\theta}}
              \\frac{d\\sigma_{\\theta}}{dz}\\right)^{2}

    Parameters
    ----------
    n2 : array_like
         Brünt-Väisälä Frequency squared [s :sup:`-2`]

    Returns
    -------
    n : array_like
        Brünt-Väisälä Frequency not-squared [s :sup:`-1`]

    Examples
    --------
    >>> import numpy as np
    >>> import oceans.sw_extras.sw_extras as swe
    >>> s = np.array([[0, 0, 0], [15, 15, 15], [30, 30, 30],[35,35,35]])
    >>> t = np.repeat(15, s.size).reshape(s.shape)
    >>> p = [[0], [250], [500], [1000]]
    >>> lat = [30,32,35]
    >>> swe.N(sw.bfrq(s, t, p, lat)[0])
    array([[0.02124956, 0.02125302, 0.02125843],
           [0.02110919, 0.02111263, 0.02111801],
           [0.00860812, 0.00860952, 0.00861171]])


    References
    ----------
    A.E. Gill 1982. p.54  eqn 3.7.15 "Atmosphere-Ocean Dynamics"
    Academic Press: New York. ISBN: 0-12-283522-0

    Jackett, David R., Trevor J. Mcdougall, 1995: Minimal Adjustment of
    Hydrographic Profiles to Achieve Static Stability. J. Atmos. Oceanic
    Technol., 12, 381-389. doi: 10.1175/1520-0426(1995)012<0381:MAOHPT>2.0.CO;2

    """
    bvfr2 = np.asanyarray(bvfr2)
    return np.sqrt(np.abs(bvfr2)) * np.sign(bvfr2)


def cph(bvfr2):
    """
    Buoyancy frequency in Cycles Per Hour.

    Parameters
    ----------
    n2 : array_like
         Brünt-Väisälä Frequency squared [s :sup:`-2`]

    Returns
    -------
    cph : array_like
          Brünt-Väisälä Frequency [ cylcles hour :sup:`-1`]

    Examples
    --------
    >>> import numpy as np
    >>> import oceans.sw_extras.sw_extras as swe
    >>> s = np.array([[0, 0, 0], [15, 15, 15], [30, 30, 30],[35,35,35]])
    >>> t = np.repeat(15, s.size).reshape(s.shape)
    >>> p = [[0], [250], [500], [1000]]
    >>> lat = [30,32,35]
    >>> swe.cph(sw.bfrq(s, t, p, lat)[0])
    array([[12.17509899, 12.17708145, 12.18018192],
           [12.09467754, 12.09664676, 12.09972655],
           [ 4.93208775,  4.9328907 ,  4.93414649]])

    References
    ----------
    A.E. Gill 1982. p.54  eqn 3.7.15 "Atmosphere-Ocean Dynamics"
    Academic Press: New York. ISBN: 0-12-283522-0

    """
    bvfr2 = np.asanyarray(bvfr2)

    # Root squared preserving the sign.
    bvfr = np.sqrt(np.abs(bvfr2)) * np.sign(bvfr2)
    return bvfr * 60.0 * 60.0 / (2.0 * np.pi)


def shear(z, u, v=0):
    r"""
    Calculates the vertical shear for u, v velocity section.

    .. math::
        \\textrm{shear} = \\frac{\\partial (u^2 + v^2)^{0.5}}{\partial z}

    Parameters
    ----------
    z : array_like
        depth [m]
    u(z) : array_like
           Eastward velocity [m s :sup:`-1`]
    v(z) : array_like
           Northward velocity [m s :sup:`-1`]

    Returns
    -------
    shr : array_like
          frequency [s :sup:`-1`]
    z_ave : array_like
            depth between z grid (M-1xN)  [m]

    Examples
    --------
    >>> import oceans.sw_extras.sw_extras as swe
    >>> z = [[0], [250], [500], [1000]]
    >>> u = [[0.5, 0.5, 0.5], [0.15, 0.15, 0.15],
    ...      [0.03, 0.03, .03], [0.,0.,0.]]
    >>> swe.shear(z, u)[0]
    array([[-1.4e-03, -1.4e-03, -1.4e-03],
           [-4.8e-04, -4.8e-04, -4.8e-04],
           [-6.0e-05, -6.0e-05, -6.0e-05]])

    """
    z, u, v = list(map(np.asanyarray, (z, u, v)))
    z, u, v = np.broadcast_arrays(z, u, v)

    m, n = z.shape
    iup = np.arange(0, m - 1)
    ilo = np.arange(1, m)
    z_ave = (z[iup, :] + z[ilo, :]) / 2.0
    vel = np.sqrt(u ** 2 + v ** 2)
    diff_vel = np.diff(vel, axis=0)
    diff_z = np.diff(z, axis=0)
    shr = diff_vel / diff_z
    return shr, z_ave


def richnumb(bvfr2, S2):
    r"""
    Calculates  the ratio of buoyancy to inertial forces which measures the
    stability of a fluid layer. this functions computes the gradient Richardson
    number in the form of:

    .. math::
        Ri = \\frac{N^2}{S^2}

    Representing a dimensionless number that expresses the ratio of the energy
    extracted by buoyancy forces to the energy gained from the shear of the
    large-scale velocity field.

    Parameters
    ----------
    bvfr2 : array_like
    Brünt-Väisälä Frequency squared (M-1xN)  [rad\ :sup:`-2` s\ :sup:`-2`]
    S2 : array_like
         shear squared [s :sup:`-2`]

    Returns
    -------
    ri : array_like
         non-dimensional

    Examples
    --------
    TODO: check the example and add real values
    >>> import numpy as np
    >>> import seawater as sw
    >>> import oceans.sw_extras.sw_extras as swe
    >>> s = np.array([[0, 0, 0], [15, 15, 15], [30, 30, 30],[ 35, 35, 35]])
    >>> t = np.repeat(15, s.size).reshape(s.shape)
    >>> p = [[0], [250], [500], [1000]]
    >>> lat = [30, 32, 35]
    >>> bvfr2 = sw.bfrq(s, t, p, lat)[0]
    >>> vel = [[0.5, 0.5, 0.5], [0.15, 0.15, 0.15],
    ...        [0.03, 0.03, .03], [0.,0.,0.]]
    >>> S2 = swe.shear(p, vel)[0] ** 2
    >>> swe.richnumb(bvfr2, S2)
    array([[  230.37941215,   230.45444299,   230.57181258],
           [ 1934.01949759,  1934.64933431,  1935.63457818],
           [20583.24410868, 20589.94661835, 20600.43125069]])


    """
    bvfr2, S2 = list(map(np.asanyarray, (bvfr2, S2)))
    # FIXME: check this for correctness.
    return bvfr2 / S2


def cor_beta(lat):
    """
    Calculates the Coriolis :math:`\\beta` factor defined by:

    .. math::
        beta = 2 \\Omega \\cos(lat)

    where:

    .. math::
        \\Omega = \\frac{2 \\pi}{\\textrm{sidereal day}} = 7.292e^{-5}
        \\textrm{ radians sec}^{-1}

    Parameters
    ----------
    lat : array_like
          latitude in decimal degrees north [-90..+90].

    Returns
    -------
    beta : array_like
        Beta Coriolis [s :sup:`-1`]

    Examples
    --------
    >>> import oceans.sw_extras.sw_extras as swe
    >>> swe.cor_beta(0)
    2.2891225867210798e-11

    References
    ----------
    S. Pond & G.Pickard 2nd Edition 1986 Introductory Dynamical
    Oceanography Pergamon Press Sydney. ISBN 0-08-028728-X

    A.E. Gill 1982. p.54  eqn 3.7.15 "Atmosphere-Ocean Dynamics"
    Academic Press: New York. ISBN: 0-12-283522-0

    """
    lat = np.asanyarray(lat)
    return 2 * OMEGA * np.cos(lat) / earth_radius


def inertial_period(lat):
    """
    Calculate the inertial period as:

    .. math::
        Ti = \\frac{2\\pi}{f} = \\frac{T_{sd}}{2\\sin\\phi}

    Parameters
    ----------
    lat : array_like
          latitude in decimal degrees north [-90..+90]

    Returns
    -------
    Ti : array_like
         period in seconds

    Examples
    --------
    >>> import oceans.sw_extras.sw_extras as swe
    >>> lat = 30.
    >>> swe.inertial_period(lat)/3600
    23.93484986278565

    """
    lat = np.asanyarray(lat)
    return 2 * np.pi / sw.f(lat)


def strat_period(N):
    """
    Stratification period is the inverse of the Buoyancy frequency and it
    is defined by:

    .. math:: Tn = \\frac{2\\pi}{N}

    Parameters
    ----------
    N : array_like
        Brünt-Väisälä Frequency [s :sup:`-1`]

    Returns
    -------
    Tn : array_like
         Brünt-Väisälä Period [s]

    Examples
    --------
    >>> import numpy as np
    >>> import seawater as sw
    >>> import oceans.sw_extras.sw_extras as swe
    >>> s = np.array([[0, 0, 0], [15, 15, 15], [30, 30, 30],[35,35,35]])
    >>> t = np.repeat(15, s.size).reshape(s.shape)
    >>> p = [[0], [250], [500], [1000]]
    >>> lat = [30,32,35]
    >>> swe.strat_period(swe.N( sw.bfrq(s, t, p, lat)[0]))
    array([[295.68548089, 295.63734267, 295.56208791],
           [297.6515901 , 297.60313502, 297.52738493],
           [729.91402019, 729.79520847, 729.60946944]])

    """
    N = np.asanyarray(N)
    return 2 * np.pi / N


def visc(s, t, p):
    """
    Calculates kinematic viscosity of sea-water.  Based on Dan Kelley's fit
    to Knauss's TABLE II-8.

    Parameters
    ----------
    s : array_like
        salinity [psu (PSS-78)]
    t : array_like
        temperature [℃ (ITS-90)]  # FIXME: [degree C (IPTS-68)]
    p : array_like
        pressure [db]

    Returns
    -------
    visc : kinematic viscosity of sea-water [m^2/s]

    Notes
    -----
    From matlab airsea

    Examples
    --------
    >>> import oceans.sw_extras.sw_extras as swe
    >>> swe.visc(40., 40., 1000.)
    8.200192496633804e-07

    Modifications: Original 1998/01/19 - Ayal Anis 1998

    """
    s, t, p = np.broadcast_arrays(s, t, p)

    visc = 1e-4 * (17.91 - 0.5381 * t + 0.00694 * t ** 2 + 0.02305 * s)
    visc /= sw.dens(s, t, p)

    return visc


def tcond(s, t, p):
    """
    Calculates thermal conductivity of sea-water.

    Parameters
    ----------
    s(p) : array_like
           salinity [psu (PSS-78)]
    t(p) : array_like
           temperature [:math:`^\\circ` C (ITS-90)]
    p : array_like
        pressure [db]

    Returns
    -------
    therm : array_like
           thermal conductivity [W m :sup: `-1` K :sup: `-1`]

    Notes
    -----
    From matlab airsea

    Examples
    --------
    >>> import oceans.sw_extras.sw_extras as swe
    >>> swe.tcond(35, 20, 0)
    0.5972445569999999

    References
    ----------
    Caldwell's DSR 21:131-137 (1974)  eq. 9
    Catelli et al.'s DSR 21:311-3179(1974)  eq. 5

    Modifications: Original 1998/01/19 - Ayal Anis 1998

    """
    s, t, p = list(map(np.asanyarray, (s, t, p)))

    if False:  # Castelli's option.
        therm = 100.0 * (
            5.5286e-3 + 3.4025e-8 * p + 1.8364e-5 * t - 3.3058e-9 * t ** 3
        )  # [W/m/K]

    # 1) Caldwell's option # 2 - simplified formula, accurate to 0.5% (eqn. 9)
    # in [cal/cm/C/sec]
    therm = 0.001365 * (
        1.0 + 0.003 * t - 1.025e-5 * t ** 2 + 0.0653 * (1e-4 * p) - 0.00029 * s
    )
    return therm * 418.4  # [cal/cm/C/sec] ->[ W/m/K]


def spice(s, t, p):
    r"""
    Compute sea spiciness as defined by Flament (2002).

    .. math:: \pi(\theta,s) = \sum^5_{i=0} \sum^4_{j=0} b_{ij}\theta^i(s-35)^i

    Parameters
    ----------
    s(p) : array_like
           salinity [psu (PSS-78)]
    t(p) : array_like
           temperature [:math:`^\\circ` C (ITS-90)]
    p : array_like
        pressure [db]

    Returns
    -------
    sp : array_like
         :math:`\pi` [kg m :sup:`3`]

    See Also
    --------
    pressure is not used... should the input be theta instead of t?
    Go read the paper!

    Notes
    -----
    Spiciness, just like potential density, is only useful over limited
    vertical excursions near the pressure to which they are referenced; for
    large vertical ranges, the slope of the isopycnals and spiciness isopleths
    vary signiﬁcantly with pressure, and generalization of the polynomial
    expansion to include a reference pressure dependence is needed.

    Examples
    --------
    >>> import oceans.sw_extras.sw_extras as swe
    >>> swe.spice(33, 15, 0)
    array(0.54458641)

    References
    ----------
    A state variable for characterizing water masses and their
    diffusive stability: spiciness. Prog. in Oceanography Volume 54, 2002,
    Pages 493-501.

    http://www.satlab.hawaii.edu/spice/spice.m

    """
    s, t, p = list(map(np.asanyarray, (s, t, p)))
    # FIXME: I'm not sure about this next step.
    pt = sw.ptmp(s, t, p)

    B = np.zeros((6, 5))
    B[0, 0] = 0.0
    B[0, 1] = 7.7442e-001
    B[0, 2] = -5.85e-003
    B[0, 3] = -9.84e-004
    B[0, 4] = -2.06e-004

    B[1, 0] = 5.1655e-002
    B[1, 1] = 2.034e-003
    B[1, 2] = -2.742e-004
    B[1, 3] = -8.5e-006
    B[1, 4] = 1.36e-005

    B[2, 0] = 6.64783e-003
    B[2, 1] = -2.4681e-004
    B[2, 2] = -1.428e-005
    B[2, 3] = 3.337e-005
    B[2, 4] = 7.894e-006

    B[3, 0] = -5.4023e-005
    B[3, 1] = 7.326e-006
    B[3, 2] = 7.0036e-006
    B[3, 3] = -3.0412e-006
    B[3, 4] = -1.0853e-006

    B[4, 0] = 3.949e-007
    B[4, 1] = -3.029e-008
    B[4, 2] = -3.8209e-007
    B[4, 3] = 1.0012e-007
    B[4, 4] = 4.7133e-008

    B[5, 0] = -6.36e-010
    B[5, 1] = -1.309e-009
    B[5, 2] = 6.048e-009
    B[5, 3] = -1.1409e-009
    B[5, 4] = -6.676e-010

    sp = np.zeros_like(pt)
    T = np.ones_like(pt)
    s = s - 35
    r, c = B.shape
    for i in range(r):
        S = np.ones_like(pt)
        for j in range(c):
            sp += B[i, j] * T * S
            S *= s
        T *= pt

    return sp


def psu2ppt(psu):
    """
    Converts salinity from PSU units to PPT
    http://stommel.tamu.edu/~baum/paleo/ocean/node31.html#PracticalSalinityScale

    """

    a = [0.008, -0.1692, 25.3851, 14.0941, -7.0261, 2.7081]
    return (
        a[1]
        + a[2] * psu ** 0.5
        + a[3] * psu
        + a[4] * psu ** 1.5
        + a[5] * psu ** 2
        + a[6] * psu ** 2.5
    )


def soundspeed(S, T, D, equation="mackenzie"):
    """
    Various sound-speed equations.
    1)  soundspeed(s, t, d) returns the sound speed (m/sec) given vectors
       of salinity (ppt), temperature (deg C) and DEPTH (m) using
       the formula of Mackenzie:  Mackenzie, K.V. "Nine-term Equation for
       Sound Speed in the Oceans", J. Acoust. Soc. Am. 70 (1981), 807-812.

    2) soundspeed(s, t, p, 'del_grosso') returns the sound speed (m/sec)given
       vectors of salinity (ppt), temperature (deg C), and  PRESSURE (dbar)
       using the Del Grosso equation:  Del Grosso, "A New Equation for the
       speed of sound in Natural Waters", J. Acoust. Soc. Am. 56#4 (1974).

    3) soundspeed(s, t, p, 'chen') returns the sound speed (m/sec) given
       vectors of salinity (ppt), temperature (deg C), and PRESSURE (dbar)
       using the Chen and Millero equation:  Chen and Millero, "The Sound
       Speed in Seawater", J. Acoust. Soc. Am. 62 (1977), 1129-1135.

    4) soundspeed(s, t, p, 'state') returns the sound speed (m/sec) given
       vectors of salinity (ppt), temperature (deg C), and PRESSURE (dbar) by
       using derivatives of the EOS80 equation of state for seawater and the
       adiabatic lapse rate.

    Notes: RP (WHOI) 3/dec/91
            Added state equation ss

    """
    if equation == "mackenzie":
        c = 1.44896e3
        t = 4.591e0
        t2 = -5.304e-2
        t3 = 2.374e-4
        s = 1.340e0
        d = 1.630e-2
        d2 = 1.675e-7
        ts = -1.025e-2
        td3 = -7.139e-13
        ssp = (
            c
            + t * T
            + t2 * T * T
            + t3 * T * T * T
            + s * (S - 35.0)
            + d * D
            + d2 * D * D
            + ts * T * (S - 35.0)
            + td3 * T * D * D * D
        )
    elif equation == "del_grosso":
        # Del grosso uses pressure in kg/cm^2.  To get to this from dbars
        # we  must divide by "g".  From the UNESCO algorithms (referring to
        # ANON (1970) BULLETIN GEODESIQUE) we have this formula for g as a
        # function of latitude and pressure.  We set latitude to 45 degrees
        # for convenience!
        XX = np.sin(45 * np.pi / 180)
        GR = 9.780318 * (1.0 + (5.2788e-3 + 2.36e-5 * XX) * XX) + 1.092e-6 * D
        P = D / GR
        # This is from VSOUND.f.
        C000 = 1402.392
        DCT = (
            0.501109398873e1 - (0.550946843172e-1 - 0.221535969240e-3 * T) * T
        ) * T
        DCS = (0.132952290781e1 + 0.128955756844e-3 * S) * S
        DCP = (
            0.156059257041e0 + (0.244998688441e-4 - 0.883392332513e-8 * P) * P
        ) * P
        DCSTP = (
            (
                -0.127562783426e-1 * T * S
                + 0.635191613389e-2 * T * P
                + 0.265484716608e-7 * T * T * P * P
                - 0.159349479045e-5 * T * P * P
                + 0.522116437235e-9 * T * P * P * P
                - 0.438031096213e-6 * T * T * T * P
            )
            - 0.161674495909e-8 * S * S * P * P
            + 0.968403156410e-4 * T * T * S
            + 0.485639620015e-5 * T * S * S * P
            - 0.340597039004e-3 * T * S * P
        )
        ssp = C000 + DCT + DCS + DCP + DCSTP
    elif equation == "chen":
        P0 = D
        # This is copied directly from the UNESCO algorithms.
        # CHECKVALUE: SVEL=1731.995 M/S, S=40 (IPSS-78),T=40 DEG C,P=10000 DBAR
        # SCALE PRESSURE TO BARS
        P = P0 / 10.0
        SR = np.sqrt(np.abs(S))
        # S**2 TERM.
        D = 1.727e-3 - 7.9836e-6 * P
        # S**3/2 TERM.
        B1 = 7.3637e-5 + 1.7945e-7 * T
        B0 = -1.922e-2 - 4.42e-5 * T
        B = B0 + B1 * P
        # S**1 TERM.
        A3 = (-3.389e-13 * T + 6.649e-12) * T + 1.100e-10
        A2 = ((7.988e-12 * T - 1.6002e-10) * T + 9.1041e-9) * T - 3.9064e-7
        A1 = (
            ((-2.0122e-10 * T + 1.0507e-8) * T - 6.4885e-8) * T - 1.2580e-5
        ) * T + 9.4742e-5
        A0 = (
            ((-3.21e-8 * T + 2.006e-6) * T + 7.164e-5) * T - 1.262e-2
        ) * T + 1.389
        A = ((A3 * P + A2) * P + A1) * P + A0
        # S**0 TERM.
        C3 = (-2.3643e-12 * T + 3.8504e-10) * T - 9.7729e-9
        C2 = (
            ((1.0405e-12 * T - 2.5335e-10) * T + 2.5974e-8) * T - 1.7107e-6
        ) * T + 3.1260e-5
        C1 = (
            ((-6.1185e-10 * T + 1.3621e-7) * T - 8.1788e-6) * T + 6.8982e-4
        ) * T + 0.153563
        C0 = (
            (((3.1464e-9 * T - 1.47800e-6) * T + 3.3420e-4) * T - 5.80852e-2)
            * T
            + 5.03711
        ) * T + 1402.388
        C = ((C3 * P + C2) * P + C1) * P + C0
        # SOUND SPEED RETURN.
        ssp = C + (A + B * SR + D * S) * S
    else:
        raise TypeError(
            "Unrecognizable equation specified: {}".format(equation)
        )
    return ssp


def photic_depth(z, par):
    """
    Computes photic depth, based on 1% of surface PAR (Photosynthetically
    Available Radiation).

    Parameters
    ----------
    z : array_like
        depth in meters.
    par : array_like
        float values of PAR

    Returns
    -------
    photic_depth : array_like
        Array of depth in which light is available.
    photic_ix : array_like
        Index of available `par` data from surface to critical depth

    """
    photic_ix = np.where(par >= par[0] / 100.0)[0]
    photic_depth = z[photic_ix]
    return photic_depth, photic_ix


def cr_depth(z, par):
    """
    Computes Critical depth. Depth where 1% of surface PAR (Photosynthetically
    Available Radiation).

    Parameters
    ----------
    z : array_like
        depth in meters.
    par : array_like
        float values of PAR

    Returns
    -------
    crdepth : int
        Critical depth. Depth where 1% of surface PAR is available.

    """
    ix = photic_depth(z, par)[1]
    crdepth = z[ix][-1]
    return crdepth


def kdpar(z, par, boundary):
    """
    Compute Kd value, since light extinction coefficient can be computed
    from depth and Photossintetically Available Radiation (PAR).
    It will compute a linear regression through out following depths from
    boundary and them will be regressed to the upper depth to boundary
    limits.

    Parameters
    ----------
    z : array_like
        depth in meters of respective PAR values
    par : array_like
        PAR values
    boundary : np.float
        First good upper limit of downcast, when PAR data has stabilized

    Return
    ------
    kd : float
        Light extinction coefficient.
    par_surface : float
        Surface PAR, modeled from first meters data.

    References
    ----------
    Smith RC, Baker KS (1978) Optical classification of natural waters.
        Limnol Ocenogr 23:260-267.

    """
    # First linear regression. Returns fit parameters to be used on
    # reconstruction of surface PAR.
    b = np.int32(boundary)
    i_b = np.where(z <= b)[0]
    par_b = par[i_b]
    z_b = z[i_b]
    z_light = photic_depth(z_b, par_b)[1]
    par_z = par_b[z_light]
    z_z = z_b[z_light]
    xp = np.polyfit(z_z, np.log(par_z), 1)

    # Linear regression based on surface PAR, obtained from linear fitting.
    # z = 0
    # PAR_surface = a(z) + b
    par_surface = np.exp(xp[1])
    par = np.r_[par_surface, par]
    z = np.r_[0, z]
    z_par = photic_depth(z, par)[1]
    kd = (np.log(par[0]) - np.log(par[b])) / z_par[b]

    return kd, par_surface


def zmld_so(s, t, p, threshold=0.05, smooth=None):
    """
    Computes mixed layer depth of Southern Ocean waters.

    Parameters
    ----------
    s : array_like
        salinity [psu (PSS-78)]
    t : array_like
        temperature [℃ (ITS-90)]
    p : array_like
        pressure [db].
    smooth : int
        size of running mean window, to smooth data.

    References
    ----------
    Mitchell B. G., Holm-Hansen, O., 1991. Observations of modeling of the
        Antartic phytoplankton crop in relation to mixing depth. Deep Sea
        Research, 38(89):981-1007. doi:10.1016/0198-0149(91)90093-U

    """
    from pandas import rolling

    sigma_t = sigmatheta(s, t, p)
    depth = copy(p)
    if smooth is not None:
        sigma_t = rolling(sigma_t, smooth, min_periods=1).mean()

    sublayer = np.where(depth[(depth >= 5) & (depth <= 10)])[0]
    sigma_x = np.nanmean(sigma_t[sublayer])
    nan_sigma = np.where(sigma_t < sigma_x + threshold)[0]
    sigma_t[nan_sigma] = np.nan
    der = np.divide(np.diff(sigma_t), np.diff(depth))
    mld = np.where(der == np.nanmax(der))[0]
    zmld = depth[mld]

    return zmld


def zmld_boyer(s, t, p):
    """
    Computes mixed layer depth, based on de Boyer Montégut et al., 2004.

    Parameters
    ----------
    s : array_like
        salinity [psu (PSS-78)]
    t : array_like
        temperature [℃ (ITS-90)]
    p : array_like
        pressure [db].

    Notes
    -----
    Based on density with fixed threshold criteria
    de Boyer Montégut et al., 2004. Mixed layer depth over the global ocean:
        An examination of profile data and a profile-based climatology.
        doi:10.1029/2004JC002378

    dataset for test and more explanation can be found at:
    http://www.ifremer.fr/cerweb/deboyer/mld/Surface_Mixed_Layer_Depth.php

    Codes based on : http://mixedlayer.ucsd.edu/

    """
    m = len(np.nonzero(~np.isnan(s))[0])

    if m <= 1:
        mldepthdens_mldindex = 0
        mldepthptemp_mldindex = 0
        return mldepthdens_mldindex, mldepthptemp_mldindex
    else:
        # starti = min(find((pres-10).^2==min((pres-10).^2)));
        starti = np.min(
            np.where(((p - 10.0) ** 2 == np.min((p - 10.0) ** 2)))[0]
        )
        starti = 0
        pres = p[starti:m]
        sal = s[starti:m]
        temp = t[starti:m]

        pden = sw.dens0(sal, temp) - 1000

        mldepthdens_mldindex = m - 1
        for i, pp in enumerate(pden):
            if np.abs(pden[starti] - pp) > 0.03:
                mldepthdens_mldindex = i
                break

        # Interpolate to exactly match the potential density threshold.
        presseg = [pres[mldepthdens_mldindex - 1], pres[mldepthdens_mldindex]]
        pdenseg = [
            pden[starti] - pden[mldepthdens_mldindex - 1],
            pden[starti] - pden[mldepthdens_mldindex],
        ]
        P = np.polyfit(presseg, pdenseg, 1)
        presinterp = np.linspace(presseg[0], presseg[1], 3)
        pdenthreshold = np.polyval(P, presinterp)

        # The potential density threshold MLD value:
        ix = np.max(np.where(np.abs(pdenthreshold) < 0.03)[0])
        mldepthdens_mldindex = presinterp[ix]

        # Search for the first level that exceeds the temperature threshold.
        mldepthptmp_mldindex = m - 1
        for i, tt in enumerate(temp):
            if np.abs(temp[starti] - tt) > 0.2:
                mldepthptmp_mldindex = i
                break

        # Interpolate to exactly match the temperature threshold.
        presseg = [pres[mldepthptmp_mldindex - 1], pres[mldepthptmp_mldindex]]
        tempseg = [
            temp[starti] - temp[mldepthptmp_mldindex - 1],
            temp[starti] - temp[mldepthptmp_mldindex],
        ]
        P = np.polyfit(presseg, tempseg, 1)
        presinterp = np.linspace(presseg[0], presseg[1], 3)
        tempthreshold = np.polyval(P, presinterp)

        # The temperature threshold MLD value:
        ix = np.max(np.where(np.abs(tempthreshold) < 0.2)[0])
        mldepthptemp_mldindex = presinterp[ix]

        return mldepthdens_mldindex, mldepthptemp_mldindex


def o2sol_SP_pt_benson_krause_84(SP, pt):
    """
    Calculates the oxygen, O2, concentration expected at equilibrium with air
    at an Absolute Pressure of 101325 Pa (sea pressure of 0 dbar) including
    saturated water vapor.

    This function uses the solubility coefficients derived from the data of
    Benson and Krause 1984, as fitted by Garcia and Gordon 1992.

    Better in the range:
      tF >= t >= 40 degC
      0 >= t >= 42 %o.

    Parameters
    ----------
    SP : array_like
        Practical Salinity
    pt : array_like
        Potential temperature [℃ (ITS-90)]

    Examples
    --------
    >>> SP = [34.7118, 34.8915, 35.0256, 34.8472, 34.7366, 34.7324]
    >>> pt = [28.8099, 28.4392, 22.7862, 10.2262, 6.8272, 4.3236]
    >>> o2sol = o2sol_SP_pt_benson_krause_84(SP, pt)
    >>> expected = [194.68254317, 195.61350628, 214.65593602, 273.56528327, 295.15807614, 312.95987166]
    >>> np.testing.assert_almost_equal(expected, o2sol)


    https://aslopubs.onlinelibrary.wiley.com/doi/pdf/10.4319/lo.1992.37.6.1307

    """
    SP, pt = list(map(np.asanyarray, (SP, pt)))

    S = SP  # rename to make eq. identical to the paper and increase readability.
    pt68 = pt * 1.00024  # IPTS-68 potential temperature in degC.

    Ts = np.log((298.15 - pt68) / (273.15 + pt68))

    # The coefficents for Benson and Krause 1984
    # from the table 1 of Garcia and Gordon (1992).
    A = [5.80871, 3.20291, 4.17887, 5.10006, -9.86643e-2, 3.80369]
    B = [-7.01577e-3, -7.70028e-3, -1.13864e-2, -9.51519e-3]
    C0 = -2.75915e-7

    # Equation 8 from Garcia and Gordon 1992 accoring to Pilson.
    lnCo = (
        A[0]
        + A[1] * Ts
        + A[2] * Ts ** 2
        + A[3] * Ts ** 3
        + A[4] * Ts ** 4
        + A[5] * Ts ** 5
        + S * (B[0] + B[1] * Ts + B[2] * Ts ** 2 + B[3] * Ts ** 3)
        + C0 * S ** 2
    )
    return np.exp(lnCo)



import numpy as np


def lanc(numwt, haf):
    """
    Generates a numwt + 1 + numwt lanczos cosine low pass filter with -6dB
    (1/4 power, 1/2 amplitude) point at haf

    Parameters
    ----------
    numwt : int
            number of points
    haf : float
          frequency (in 'cpi' of -6dB point, 'cpi' is cycles per interval.
          For hourly data cpi is cph,

    Examples
    --------
    >>> from oceans.filters import lanc
    >>> import matplotlib.pyplot as plt
    >>> t = np.arange(500)  # Time in hours.
    >>> h = 2.5 * np.sin(2 * np.pi * t / 12.42)
    >>> h += 1.5 * np.sin(2 * np.pi * t / 12.0)
    >>> h += 0.3 * np.random.randn(len(t))
    >>> wt = lanc(96+1+96, 1./40)
    >>> low = np.convolve(wt, h, mode='same')
    >>> high = h - low
    >>> fig, (ax0, ax1) = plt.subplots(nrows=2)
    >>> _ = ax0.plot(high, label='high')
    >>> _ = ax1.plot(low, label='low')
    >>> _ = ax0.legend(numpoints=1)
    >>> _ = ax1.legend(numpoints=1)

    """
    summ = 0
    numwt += 1
    wt = np.zeros(numwt)

    # Filter weights.
    ii = np.arange(numwt)
    wt = 0.5 * (1.0 + np.cos(np.pi * ii * 1.0 / numwt))
    ii = np.arange(1, numwt)
    xx = np.pi * 2 * haf * ii
    wt[1 : numwt + 1] = wt[1 : numwt + 1] * np.sin(xx) / xx
    summ = wt[1 : numwt + 1].sum()
    xx = wt.sum() + summ
    wt /= xx
    return np.r_[wt[::-1], wt[1 : numwt + 1]]


def smoo1(datain, window_len=11, window="hanning"):
    """
    Smooth the data using a window with requested size.

    Parameters
    ----------
    datain : array_like
             input series
    window_len : int
                 size of the smoothing window; should be an odd integer
    window : str
             window from 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'.
             flat window will produce a moving average smoothing.

    Returns
    -------
    data_out : array_like
            smoothed signal

    See Also
    --------
    scipy.signal.lfilter

    Notes
    -----
    original from: https://scipy-cookbook.readthedocs.io/items/SignalSmooth.html
    This method is based on the convolution of a scaled window with the signal.
    The signal is prepared by introducing reflected copies of the signal (with
    the window size) in both ends so that transient parts are minimized in the
    beginning and end part of the output signal.

    Examples
    --------
    >>> from oceans.filters import smoo1
    >>> import numpy as np
    >>> import matplotlib.pyplot as plt
    >>> time = np.linspace( -4, 4, 100 )
    >>> series = np.sin(time)
    >>> noise_series = series + np.random.randn( len(time) ) * 0.1
    >>> data_out = smoo1(series)
    >>> ws = 31
    >>> ax = plt.subplot(211)
    >>> _ = ax.plot(np.ones(ws))
    >>> windows = ['flat', 'hanning', 'hamming', 'bartlett', 'blackman']
    >>> for w in windows[1:]:
    ...     _ = eval('plt.plot(np.' + w + '(ws) )')
    >>> _ = ax.axis([0, 30, 0, 1.1])
    >>> leg = ax.legend(windows)
    >>> _ = plt.title('The smoothing windows')
    >>> ax = plt.subplot(212)
    >>> l1, = ax.plot(series)
    >>> l2, = ax.plot(noise_series)
    >>> for w in windows:
    ...     _ = plt.plot(smoo1(noise_series, 10, w))
    >>> l = ['original signal', 'signal with noise']
    >>> l.extend(windows)
    >>> leg = ax.legend(l)
    >>> _ = plt.title('Smoothing a noisy signal')

    TODO: window parameter can be the window itself (i.e. an array)
    instead of a string.

    """

    datain = np.asarray(datain)

    if datain.ndim != 1:
        raise ValueError("Smooth only accepts 1 dimension arrays.")

    if datain.size < window_len:
        raise ValueError("Input vector needs to be bigger than window size.")

    if window_len < 3:
        return datain

    if window not in ["flat", "hanning", "hamming", "bartlett", "blackman"]:
        msg = "Window must be is one of 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'"  # noqa
        raise ValueError(msg)

    s = np.r_[
        2 * datain[0] - datain[window_len:1:-1],
        datain,
        2 * datain[-1] - datain[-1:-window_len:-1],
    ]

    if window == "flat":  # Moving average.
        w = np.ones(window_len, "d")
    else:
        w = eval("np." + window + "(window_len)")

    data_out = np.convolve(w / w.sum(), s, mode="same")
    return data_out[window_len - 1 : -window_len + 1]


def smoo2(A, hei, wid, kind="hann", badflag=-9999, beta=14):
    """
    Calculates the smoothed array 'As' from the original array 'A' using the
    specified window of type 'kind' and shape ('hei', 'wid').

    Usage:
    As = smoo2(A, hei, wid, kind='hann', badflag=-9999, beta=14)

    Parameters
    ----------
    A : 2D array
      Array to be smoothed.
    hei : integer
      Window height. Must be odd and greater than or equal to 3.
    wid : integer
      Window width. Must be odd and greater than or equal to 3.
    kind : string, optional
      Refer to Numpy for details about each window type.
    badflag : float, optional
      The bad data flag. Elements of the input array 'A' holding this value are ignored.
    beta : float, optional
      Shape parameter for the kaiser window.

    Returns
    -------
    As : 2D array
      The smoothed array.

    André Palóczy Filho (paloczy@gmail.com)
    April 2012

    """
    # Checking window type and dimensions
    kinds = ["hann", "hamming", "blackman", "bartlett", "kaiser"]
    if kind not in kinds:
        raise ValueError("Invalid window type requested: %s" % kind)

    if (np.mod(hei, 2) == 0) or (np.mod(wid, 2) == 0):
        raise ValueError("Window dimensions must be odd")

    if (hei <= 1) or (wid <= 1):
        raise ValueError("Window shape must be (3,3) or greater")

    # Creating the 2D window.
    if kind == "kaiser":  # If the window kind is kaiser (beta is required).
        wstr = "np.outer(np.kaiser(hei, beta), np.kaiser(wid, beta))"
    # If the window kind is hann, hamming, blackman or bartlett
    # (beta is not required).
    else:
        if kind == "hann":
            # Converting the correct window name (Hann) to the numpy function
            # name (numpy.hanning).
            kind = "hanning"
        # Computing outer product to make a 2D window out of the original 1d
        # windows.
        # TODO: Get rid of this evil eval.
        wstr = "np.outer(np." + kind + "(hei), np." + kind + "(wid))"
    wdw = eval(wstr)

    A = np.asanyarray(A)
    Fnan = np.isnan(A)
    imax, jmax = A.shape
    As = np.NaN * np.ones((imax, jmax))

    for i in range(imax):
        for j in range(jmax):
            # Default window parameters.
            wupp = 0
            wlow = hei
            wlef = 0
            wrig = wid
            lh = np.floor(hei / 2)
            lw = np.floor(wid / 2)

            # Default array ranges (functions of the i, j indices).
            upp = i - lh
            low = i + lh + 1
            lef = j - lw
            rig = j + lw + 1

            # Tiling window and input array at the edges.
            # Upper edge.
            if upp < 0:
                wupp = wupp - upp
                upp = 0

            # Left edge.
            if lef < 0:
                wlef = wlef - lef
                lef = 0

            # Bottom edge.
            if low > imax:
                ex = low - imax
                wlow = wlow - ex
                low = imax

            # Right edge.
            if rig > jmax:
                ex = rig - jmax
                wrig = wrig - ex
                rig = jmax

            # Computing smoothed value at point (i, j).
            Ac = A[upp:low, lef:rig]
            wdwc = wdw[wupp:wlow, wlef:wrig]
            fnan = np.isnan(Ac)
            Ac[fnan] = 0
            wdwc[fnan] = 0  # Eliminating NaNs from mean computation.
            fbad = Ac == badflag
            wdwc[fbad] = 0  # Eliminating bad data from mean computation.
            a = Ac * wdwc
            As[i, j] = a.sum() / wdwc.sum()
    # Assigning NaN to the positions holding NaNs in the original array.
    As[Fnan] = np.NaN

    return As


def weim(x, N, kind="hann", badflag=-9999, beta=14):
    """
    Calculates the smoothed array 'xs' from the original array 'x' using the
    specified window of type 'kind' and size 'N'. 'N' must be an odd number.

    Usage:
    xs = weim(x, N, kind='hann', badflag=-9999, beta=14)


    Parameters
    ----------
    x : 1D array
      Array to be smoothed.

    N : integer
      Window size. Must be odd.

    kind : string, optional
    Refer to Numpy for details about each window type.

    badflag : float, optional
              The bad data flag. Elements of the input array 'A' holding this
              value are ignored.

    beta    : float, optional
              Shape parameter for the kaiser window. For windows other than the
              kaiser window, this parameter does nothing.

    Returns
    -------
    xs      : 1D array
              The smoothed array.

    ---------------------------------------
    André Palóczy Filho (paloczy@gmail.com) June 2012

    """
    # Checking window type and dimensions.
    kinds = ["hann", "hamming", "blackman", "bartlett", "kaiser"]
    if kind not in kinds:
        raise ValueError("Invalid window type requested: %s" % kind)

    if np.mod(N, 2) == 0:
        raise ValueError("Window size must be odd")

    # Creating the window.
    if kind == "kaiser":  # If the window kind is kaiser (beta is required).
        wstr = "np.kaiser(N, beta)"
    # If the window kind is hann, hamming, blackman or bartlett (beta is not
    # required).
    else:
        if kind == "hann":
            # Converting the correct window name (Hann) to the numpy function
            # name (numpy.hanning).
            kind = "hanning"
            # Computing outer product to make a 2D window out of the original
            # 1D windows.
        wstr = "np." + kind + "(N)"

    # FIXME: Do not use `eval`.
    w = eval(wstr)
    x = np.asarray(x).flatten()
    Fnan = np.isnan(x).flatten()

    ln = (N - 1) / 2
    lx = x.size
    lf = lx - ln
    xs = np.NaN * np.ones(lx)

    # Eliminating bad data from mean computation.
    fbad = x == badflag
    x[fbad] = np.nan

    for i in range(lx):
        if i <= ln:
            xx = x[: ln + i + 1]
            ww = w[ln - i :]
        elif i >= lf:
            xx = x[i - ln :]
            ww = w[: lf - i - 1]
        else:
            xx = x[i - ln : i + ln + 1]
            ww = w.copy()

        # Counting only NON-NaNs, both in the input array and in the window
        # points.
        f = ~np.isnan(xx)
        xx = xx[f]
        ww = ww[f]

        # Thou shalt not divide by zero.
        if f.sum() == 0:
            xs[i] = x[i]
        else:
            xs[i] = np.sum(xx * ww) / np.sum(ww)

    # Assigning NaN to the positions holding NaNs in the input array.
    xs[Fnan] = np.nan

    return xs


def medfilt1(x, L=3):
    """
    Median filter for 1d arrays.

    Performs a discrete one-dimensional median filter with window length `L` to
    input vector `x`.  Produces a vector the same size as `x`.  Boundaries are
    handled by shrinking `L` at edges; no data outside of `x` is used in
    producing the median filtered output.

    Parameters
    ----------
    x : array_like
        Input 1D data
    L : integer
        Window length

    Returns
    -------
    xout : array_like
           Numpy 1d array of median filtered result; same size as x

    Examples
    --------
    >>> from oceans.filters import medfilt1
    >>> import matplotlib.pyplot as plt
    >>> # 100 pseudo-random integers ranging from 1 to 100, plus three large
    >>> # outliers for illustration.
    >>> x = np.r_[np.ceil(np.random.rand(25)*100), [1000],
    ...           np.ceil(np.random.rand(25)*100), [2000],
    ...           np.ceil(np.random.rand(25)*100), [3000],
    ...           np.ceil(np.random.rand(25)*100)]
    >>> L = 2
    >>> xout = medfilt1(x=x, L=L)
    >>> ax = plt.subplot(211)
    >>> l1, l2 = ax.plot(x), ax.plot(xout)
    >>> ax.grid(True)
    >>> y1min, y1max = np.min(xout) * 0.5, np.max(xout) * 2.0
    >>> leg1 = ax.legend(['x (pseudo-random)','xout'])
    >>> t1 = ax.set_title('''Median filter with window length %s.
    ...                   Removes outliers, tracks remaining signal)''' % L)
    >>> L = 103
    >>> xout = medfilt1(x=x, L=L)
    >>> ax = plt.subplot(212)
    >>> l1, l2, = ax.plot(x), ax.plot(xout)
    >>> ax.grid(True)
    >>> y2min, y2max = np.min(xout) * 0.5, np.max(xout) * 2.0
    >>> leg2 = ax.legend(["Same x (pseudo-random)", "xout"])
    >>> t2 = ax.set_title('''Median filter with window length %s.
    ...                   Removes outliers and noise''' % L)
    >>> ax = plt.subplot(211)
    >>> lims1 = ax.set_ylim([min(y1min, y2min), max(y1max, y2max)])
    >>> ax = plt.subplot(212)
    >>> lims2 = ax.set_ylim([min(y1min, y2min), max(y1max, y2max)])

    """

    xin = np.atleast_1d(np.asanyarray(x))
    N = len(x)
    L = int(L)  # Ensure L is odd integer so median requires no interpolation.
    if L % 2 == 0:
        L += 1

    if N < 2:
        raise ValueError("Input sequence must be >= 2.")
        return None

    if L < 2:
        raise ValueError("Input filter window length must be >=2.")
        return None

    if L > N:
        msg = "Input filter window length must be shorter than series: L = {0:d}, len(x) = {1:d}".format  # noqa
        raise ValueError(msg(L, N))
        return None

    if xin.ndim > 1:
        msg = "Input sequence has to be 1d: ndim = {}".format
        raise ValueError(msg(xin.ndim))

    xout = np.zeros_like(xin) + np.NaN

    Lwing = (L - 1) // 2

    # NOTE: Use np.ndenumerate in case I expand to +1D case
    for i, xi in enumerate(xin):
        if i < Lwing:  # Left boundary.
            xout[i] = np.median(xin[0 : i + Lwing + 1])  # (0 to i + Lwing)
        elif i >= N - Lwing:  # Right boundary.
            xout[i] = np.median(xin[i - Lwing : N])  # (i-Lwing to N-1)
        else:  # Middle (N-2*Lwing input vector and filter window overlap).
            xout[i] = np.median(xin[i - Lwing : i + Lwing + 1])
            # (i-Lwing to i+Lwing)
    return xout


def fft_lowpass(signal, low, high):
    """
    Performs a low pass filer on the series.
    low and high specifies the boundary of the filter.

    >>> from oceans.filters import fft_lowpass
    >>> import matplotlib.pyplot as plt
    >>> t = np.arange(500)  # Time in hours.
    >>> x = 2.5 * np.sin(2 * np.pi * t / 12.42)
    >>> x += 1.5 * np.sin(2 * np.pi * t / 12.0)
    >>> x += 0.3 * np.random.randn(len(t))
    >>> filtered = fft_lowpass(x, low=1/30, high=1/40)
    >>> fig, ax = plt.subplots()
    >>> l1, = ax.plot(t, x, label='original')
    >>> l2, = ax.plot(t, filtered, label='filtered')
    >>> legend = ax.legend()

    """

    if len(signal) % 2:
        result = np.fft.rfft(signal, len(signal))
    else:
        result = np.fft.rfft(signal)

    freq = np.fft.fftfreq(len(signal))[: len(signal) // 2 + 1]
    factor = np.ones_like(freq)
    factor[freq > low] = 0.0
    sl = np.logical_and(high < freq, freq < low)
    a = factor[sl]

    # Create float array of required length and reverse.
    a = np.arange(len(a) + 2).astype(float)[::-1]

    # Ramp from 1 to 0 exclusive.
    a = (a / a[0])[1:-1]

    # Insert ramp into factor.
    factor[sl] = a

    result = result * factor

    return np.fft.irfft(result, len(signal))


def md_trenberth(x):
    """
    Returns the filtered series using the Trenberth filter as described
    on Monthly Weather Review, vol. 112, No. 2, Feb 1984.

    Input data: series x of dimension 1Xn (must be at least dimension 11)
    Output data: y = md_trenberth(x) where y has dimension 1X(n-10)

    Examples
    --------
    >>> from oceans.filters import md_trenberth
    >>> import matplotlib.pyplot as plt
    >>> t = np.arange(500)  # Time in hours.
    >>> x = 2.5 * np.sin(2 * np.pi * t / 12.42)
    >>> x += 1.5 * np.sin(2 * np.pi * t / 12.0)
    >>> x += 0.3 * np.random.randn(len(t))
    >>> filtered = md_trenberth(x)
    >>> fig, ax = plt.subplots()
    >>> l1, = ax.plot(t, x, label='original')
    >>> pad = [np.NaN]*5
    >>> l2, = ax.plot(t, np.r_[pad, filtered, pad], label='filtered')
    >>> legend = ax.legend()

    """
    x = np.asanyarray(x)
    weight = np.array(
        [
            0.02700,
            0.05856,
            0.09030,
            0.11742,
            0.13567,
            0.14210,
            0.13567,
            0.11742,
            0.09030,
            0.05856,
            0.02700,
        ]
    )

    sz = len(x)
    y = np.zeros(sz - 10)
    for i in range(5, sz - 5):
        y[i - 5] = 0
        for j in range(11):
            y[i - 5] = y[i - 5] + x[i - 6 + j + 1] * weight[j]

    return y


def pl33tn(x, dt=1.0, T=33.0, mode="valid"):
    """
    Computes low-passed series from `x` using pl33 filter, with optional
    sample interval `dt` (hours) and filter half-amplitude period T (hours)
    as input for non-hourly series.

    The PL33 filter is described on p. 21, Rosenfeld (1983), WHOI
    Technical Report 85-35.  Filter half amplitude period = 33 hrs.,
    half power period = 38 hrs.  The time series x is folded over
    and cosine tapered at each end to return a filtered time series
    xf of the same length.  Assumes length of x greater than 67.

    Examples
    --------
    >>> from oceans.filters import pl33tn
    >>> import matplotlib.pyplot as plt
    >>> t = np.arange(500)  # Time in hours.
    >>> x = 2.5 * np.sin(2 * np.pi * t / 12.42)
    >>> x += 1.5 * np.sin(2 * np.pi * t / 12.0)
    >>> x += 0.3 * np.random.randn(len(t))
    >>> filtered_33 = pl33tn(x, dt=4.0)   # 33 hr filter
    >>> filtered_33d3 = pl33tn(x, dt=4.0, T=72.0)  # 3 day filter
    >>> fig, ax = plt.subplots()
    >>> l1, = ax.plot(t, x, label='original')
    >>> pad = [np.NaN]*8
    >>> l2, = ax.plot(t, np.r_[pad, filtered_33, pad], label='33 hours')
    >>> pad = [np.NaN]*17
    >>> l3, = ax.plot(t, np.r_[pad, filtered_33d3, pad], label='3 days')
    >>> legend = ax.legend()


    """

    pl33 = np.array(
        [
            -0.00027,
            -0.00114,
            -0.00211,
            -0.00317,
            -0.00427,
            -0.00537,
            -0.00641,
            -0.00735,
            -0.00811,
            -0.00864,
            -0.00887,
            -0.00872,
            -0.00816,
            -0.00714,
            -0.00560,
            -0.00355,
            -0.00097,
            +0.00213,
            +0.00574,
            +0.00980,
            +0.01425,
            +0.01902,
            +0.02400,
            +0.02911,
            +0.03423,
            +0.03923,
            +0.04399,
            +0.04842,
            +0.05237,
            +0.05576,
            +0.05850,
            +0.06051,
            +0.06174,
            +0.06215,
            +0.06174,
            +0.06051,
            +0.05850,
            +0.05576,
            +0.05237,
            +0.04842,
            +0.04399,
            +0.03923,
            +0.03423,
            +0.02911,
            +0.02400,
            +0.01902,
            +0.01425,
            +0.00980,
            +0.00574,
            +0.00213,
            -0.00097,
            -0.00355,
            -0.00560,
            -0.00714,
            -0.00816,
            -0.00872,
            -0.00887,
            -0.00864,
            -0.00811,
            -0.00735,
            -0.00641,
            -0.00537,
            -0.00427,
            -0.00317,
            -0.00211,
            -0.00114,
            -0.00027,
        ]
    )

    _dt = np.linspace(-33, 33, 67)

    dt = float(dt) * (33.0 / T)

    filter_time = np.arange(0.0, 33.0, dt, dtype="d")
    # N = len(filter_time)
    filter_time = np.hstack((-filter_time[-1:0:-1], filter_time))

    pl33 = np.interp(filter_time, _dt, pl33)
    pl33 /= pl33.sum()

    xf = np.convolve(x, pl33, mode=mode)
    return xf



import warnings

import numpy as np

from netCDF4 import Dataset

from oceans.ocfis import get_profile, wrap_lon180


def _woa_variable(variable):
    _VAR = {
        "temperature": "t",
        "salinity": "s",
        "silicate": "i",
        "phosphate": "p",
        "nitrate": "n",
        "oxygen_saturation": "O",
        "dissolved_oxygen": "o",
        "apparent_oxygen_utilization": "A",
    }
    v = _VAR.get(variable)
    if not v:
        raise ValueError(
            f'Unrecognizable variable. Expected one of {list(_VAR.keys())}, got "{variable}".'
        )
    return v


def _woa_url(variable, time_period, resolution):
    base = "https://data.nodc.noaa.gov/thredds/dodsC"

    v = _woa_variable(variable)

    if variable not in ["salinity", "temperature"]:
        pref = "woa09"
        warnings.warn(
            f'The variable "{variable}" is only available at 1 degree resolution, '
            f'annual time period, and "{pref}".'
        )
        return f"{base}/" f"{pref}/" f"{variable}_annual_1deg.nc"
    else:
        dddd = "decav"
        pref = "woa18"

    grids = {
        "5": ("5deg", "5d"),
        "1": ("1.00", "01"),
        "1/4": ("0.25", "04"),
    }
    grid = grids.get(resolution)
    if not grid:
        raise ValueError(
            f'Unrecognizable resolution. Expected one of {list(grids.keys())}, got "{resolution}".'
        )
    res = grid[0]
    gg = grid[1]

    time_periods = {
        "annual": "00",
        "january": "01",
        "february": "02",
        "march": "03",
        "april": "04",
        "may": "05",
        "june": "06",
        "july": "07",
        "august": "08",
        "september": "09",
        "october": "10",
        "november": "11",
        "december": "12",
        "winter": "13",
        "spring": "14",
        "summer": "15",
        "autumn": "16",
    }

    time_period = time_period.lower()
    if len(time_period) == 3:
        tt = [
            time_periods.get(k)
            for k in time_periods.keys()
            if k.startswith(time_period)
        ][0]
    elif len(time_period) == 2 and time_period in time_periods.values():
        tt = time_period
    else:
        tt = time_periods.get(time_period)

    if not tt:
        raise ValueError(
            f"Unrecognizable time_period. "
            f'Expected one of {list(time_periods.keys())}, got "{time_period}".'
        )

    url = (
        f"{base}/"
        "/ncei/woa/"
        f"{variable}/decav/{res}/"
        f"{pref}_{dddd}_{v}{tt}_{gg}.nc"  # '[PREF]_[DDDD]_[V][TT][FF][GG]' Is [FF] used?
    )
    return url


def woa_profile(
    lon, lat, variable="temperature", time_period="annual", resolution="1"
):
    """
    Return an iris.cube instance from a World Ocean Atlas variable at a
    given lon, lat point.

    Parameters
    ----------
    lon, lat: float
        point positions to extract the interpolated profile.
    Choose data `variable` from:
        'temperature', 'salinity', 'silicate', 'phosphate',
        'nitrate', 'oxygen_saturation', 'dissolved_oxygen', or
        'apparent_oxygen_utilization'.
    Choose `time_period` from:
        01-12: January to December
        13-16: seasonal (North Hemisphere `Winter`, `Spring`, `Summer`, and `Autumn` respectively)
        00: Annual
    Choose `resolution` from:
        '5', '1', or '1/4' degrees (str)

    Returns
    -------
    Iris.cube instance with the climatology.

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> from oceans.datasets import woa_profile
    >>> cube = woa_profile(-143, 10, variable='temperature',
    ...                    time_period='annual', resolution='5')
    >>> fig, ax = plt.subplots(figsize=(2.25, 5))
    >>> z = cube.coord(axis='Z').points
    >>> l = ax.plot(cube[0, :].data, z)
    >>> ax.grid(True)
    >>> ax.invert_yaxis()

    """
    import iris

    url = _woa_url(
        variable=variable, time_period=time_period, resolution=resolution
    )

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        cubes = iris.load_raw(url)

    # TODO: should we be using `an` instead of `mn`?
    v = _woa_variable(variable)
    cube = [c for c in cubes if c.var_name == f"{v}_mn"][0]
    scheme = iris.analysis.Nearest()
    sample_points = [("longitude", lon), ("latitude", lat)]
    kw = {
        "sample_points": sample_points,
        "scheme": scheme,
        "collapse_scalar": True,
    }
    return cube.interpolate(**kw)


def woa_subset(
    bbox,
    variable="temperature",
    time_period="annual",
    resolution="5",
    full=False,
):
    """
    Return an iris.cube instance from a World Ocean Atlas variable at a
    given lon, lat bounding box.

    Parameters
    ----------
    bbox: list, tuple
          minx, maxx, miny, maxy positions to extract.
    See `woa_profile` for the other options.

    Returns
    -------
    `iris.Cube` instance with the climatology.

    Examples
    --------
    >>> # Extract a 2D surface -- Annual temperature climatology:
    >>> import iris.plot as iplt
    >>> import matplotlib.pyplot as plt
    >>> from oceans.colormaps import cm
    >>> bbox = [2.5, 357.5, -87.5, 87.5]
    >>> cube = woa_subset(bbox, variable='temperature', time_period='annual', resolution='5')
    >>> c = cube[0, 0, ...]  # Slice singleton time and first level.
    >>> cs = iplt.pcolormesh(c, cmap=cm.avhrr)
    >>> cbar = plt.colorbar(cs)

    >>> # Extract a square around the Mariana Trench averaging into a profile.
    >>> import iris
    >>> from oceans.colormaps import get_color
    >>> colors = get_color(12)
    >>> months = 'Jan Feb Apr Mar May Jun Jul Aug Sep Oct Nov Dec'.split()
    >>> bbox = [-143, -141, 10, 12]
    >>> fig, ax = plt.subplots(figsize=(5, 5))
    >>> for month in months:
    ...     cube = woa_subset(bbox, time_period=month, variable='temperature', resolution='1')
    ...     grid_areas = iris.analysis.cartography.area_weights(cube)
    ...     c = cube.collapsed(['longitude', 'latitude'], iris.analysis.MEAN, weights=grid_areas)
    ...     z = c.coord(axis='Z').points
    ...     l = ax.plot(c[0, :].data, z, label=month, color=next(colors))
    >>> ax.grid(True)
    >>> ax.invert_yaxis()
    >>> leg = ax.legend(loc='lower left')
    >>> _ = ax.set_ylim(200, 0)

    """
    import iris

    v = _woa_variable(variable)
    url = _woa_url(variable, time_period, resolution)
    cubes = iris.load_raw(url)
    cubes = [
        cube.intersection(
            longitude=(bbox[0], bbox[1]), latitude=(bbox[2], bbox[3])
        )
        for cube in cubes
    ]

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        cubes = iris.cube.CubeList(cubes)

    if full:
        return cubes
    else:
        return [c for c in cubes if c.var_name == f"{v}_mn"][0]


def etopo_subset(bbox, tfile=None, smoo=False):
    """
    Get a etopo subset.
    Should work on any netCDF with x, y, data

    Examples
    --------
    >>> from oceans.datasets import etopo_subset
    >>> import matplotlib.pyplot as plt
    >>> bbox = [-43, -30, -22, -17]
    >>> lon, lat, bathy = etopo_subset(bbox=bbox, smoo=True)
    >>> fig, ax = plt.subplots()
    >>> cs = ax.pcolormesh(lon, lat, bathy)

    Based on trondkristiansen contourICEMaps.py
    """
    if tfile is None:
        tfile = "http://gamone.whoi.edu/thredds/dodsC/usgs/data0/bathy/ETOPO2v2c_f4.nc"

    with Dataset(tfile, "r") as etopo:
        lons = etopo.variables["x"][:]
        lats = etopo.variables["y"][:]

        imin, imax, jmin, jmax = _get_indices(bbox, lons, lats)
        lon, lat = np.meshgrid(lons[imin:imax], lats[jmin:jmax])

        # FIXME: This assumes j, i order.
        bathy = etopo.variables["z"][jmin:jmax, imin:imax]

    if smoo:
        from scipy.ndimage.filters import gaussian_filter

        bathy = gaussian_filter(bathy, sigma=1)

    return lon, lat, bathy


def get_depth(lon, lat, tfile=None):
    """
    Find the depths for each station on the etopo2 database.

    Examples
    --------
    >>> from oceans.datasets import get_depth
    >>> station_lon = [-40, -32]
    >>> station_lat = [-20, -20]
    >>> get_depth(station_lon, station_lat)
    array([  -32.98816423, -4275.63374601])

    """
    lon, lat = list(map(np.atleast_1d, (lon, lat)))

    offset = 5
    bbox = [
        lon.min() - offset,
        lon.max() + offset,
        lat.min() - offset,
        lat.max() + offset,
    ]
    lons, lats, bathy = etopo_subset(bbox, tfile=tfile, smoo=False)

    return get_profile(lons, lats, bathy, lon, lat, mode="nearest", order=3)


def get_isobath(bbox, iso=-200, tfile=None, smoo=False):
    """
    Finds an isobath on the etopo2 database and returns
    its lon, lat segments for plotting.

    Examples
    --------
    >>> from oceans.datasets import etopo_subset, get_isobath
    >>> import matplotlib.pyplot as plt
    >>> bbox = [-43, -30, -22, -17]
    >>> segments = get_isobath(bbox=bbox, iso=-200, smoo=True)
    >>> lon, lat, bathy = etopo_subset(bbox=bbox, smoo=True)
    >>> fig, ax = plt.subplots()
    >>> cs = ax.pcolormesh(lon, lat, bathy)
    >>> for segment in segments:
    ...     lines = ax.plot(segment[:, 0], segment[:, -1], 'k', linewidth=2)

    """
    import matplotlib._contour as contour

    lon, lat, topo = etopo_subset(bbox, tfile=tfile, smoo=smoo)

    # Required args for QuadContourGenerator.
    mask, corner_mask, nchunk = None, True, 0
    c = contour.QuadContourGenerator(lon, lat, topo, mask, corner_mask, nchunk)
    res = c.create_contour(iso)
    nseg = len(res) // 2
    segments = res[:nseg]
    return segments


def _minmax(v):
    return np.min(v), np.max(v)


def _get_indices(bbox, lons, lats):
    """Return the data indices for a lon, lat square."""
    lons = wrap_lon180(lons)

    idx_x = np.logical_and(lons >= bbox[0], lons <= bbox[1])
    idx_y = np.logical_and(lats >= bbox[2], lats <= bbox[3])
    if lons.ndim == 2 and lats.ndim == 2:
        inregion = np.logical_and(idx_x, idx_y)
        region_inds = np.where(inregion)
        imin, imax = _minmax(region_inds[0])
        jmin, jmax = _minmax(region_inds[1])
    elif lons.ndim == 1 and lats.ndim == 1:
        imin, imax = _minmax(np.where(idx_x))
        jmin, jmax = _minmax(np.where(idx_y))
    else:
        msg = "Cannot understand input shapes lons {!r} and lats {!r}".format
        raise ValueError(msg(lons.shape, lats.shape))
    return imin, imax + 1, jmin, jmax + 1


if __name__ == "__main__":
    import doctest

    doctest.testmod()



import re
import warnings

import gsw
import numpy as np
import numpy.ma as ma


def spdir2uv(spd, ang, deg=False):
    """
    Computes u, v components from speed and direction.

    Parameters
    ----------
    spd : array_like
          speed [m s :sup:`-1`]
    ang : array_like
          direction [deg]
    deg : bool
          option, True if data is in degrees. Default is False

    Returns
    -------
    u : array_like
        zonal wind velocity [m s :sup:`-1`]
    v : array_like
        meridional wind velocity [m s :sup:`-1`]

    """

    if deg:
        ang = np.deg2rad(ang)

    # Calculate U (E-W) and V (N-S) components
    u = spd * np.sin(ang)
    v = spd * np.cos(ang)
    return u, v


def uv2spdir(u, v, mag=0, rot=0):
    """
    Computes speed and direction from u, v components.
    Converts rectangular to polar coordinate, geographic convention
    Allows for rotation and magnetic declination correction.

    Parameters
    ----------
    u : array_like
        zonal wind velocity [m s :sup:`-1`]
    v : array_like
        meridional wind velocity [m s :sup:`-1`]
    mag : float, array_like, optional
          Magnetic correction [deg]
    rot : float, array_like
          Angle for rotation [deg]

    Returns
    -------
    ang : array_like
          direction [deg]
    spd : array_like
          speed [m s :sup:`-1`]

    Notes
    -----
    Zero degrees is north.

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> from oceans.ocfis import uv2spdir
    >>> u = [+0, -0.5, -0.50, +0.90]
    >>> v = [+1, +0.5, -0.45, -0.85]
    >>> wd, ws = uv2spdir(u,v)
    >>> fig, ax = plt.subplots(subplot_kw=dict(polar=True))
    >>> kw = dict(arrowstyle="->")
    >>> wd = np.deg2rad(wd)
    >>> lines = [ax.annotate("", xy=(d, s), xytext=(0, 0), arrowprops=kw)
    ...  for d, s in zip(wd, ws)]
    >>> _ = ax.set_ylim(0, np.max(ws))

    """

    u, v, mag, rot = list(map(np.asarray, (u, v, mag, rot)))

    vec = u + 1j * v
    spd = np.abs(vec)
    ang = np.angle(vec, deg=True)
    ang = ang - mag + rot
    ang = np.mod(90.0 - ang, 360.0)  # Zero is North.

    return ang, spd


def del_eta_del_x(U, f, g, balance="geostrophic", R=None):
    r"""
    Calculate :mat: `\frac{\partial \eta} {\partial x}` for different
    force balances

    Parameters
    ----------
    U : array_like
        velocity magnitude [m/s]
    f : float
        Coriolis parameter
    d : float
        Acceleration of gravity
    balance : str, optional
              geostrophic [default], gradient or max_gradient
    R : float, optional
        Radius

    """

    if balance == "geostrophic":
        detadx = f * U / g

    elif balance == "gradient":
        detadx = (U ** 2 / R + f * U) / g

    elif balance == "max_gradient":
        detadx = (R * f ** 2) / (4 * g)

    return detadx


def mld(SA, CT, p, criterion="pdvar"):
    r"""
    Compute the mixed layer depth.

    Parameters
    ----------
    SA : array_like
         Absolute Salinity  [g/kg]
    CT : array_like
         Conservative Temperature [:math:`^\circ` C (ITS-90)]
    p : array_like
        sea pressure [dbar]
    criterion : str, optional
               MLD Criteria

    Mixed layer depth criteria are:

    'temperature' : Computed based on constant temperature difference
    criterion, CT(0) - T[mld] = 0.5 degree C.

    'density' : computed based on the constant potential density difference
    criterion, pd[0] - pd[mld] = 0.125 in sigma units.

    'pdvar' : computed based on variable potential density criterion
    pd[0] - pd[mld] = var(T[0], S[0]), where var is a variable potential
    density difference which corresponds to constant temperature difference of
    0.5 degree C.

    Returns
    -------
    MLD : array_like
          Mixed layer depth
    idx_mld : bool array
              Boolean array in the shape of p with MLD index.

    References
    ----------
    Monterey, G., and S. Levitus, 1997: Seasonal variability of mixed
    layer depth for the World Ocean. NOAA Atlas, NESDIS 14, 100 pp.
    Washington, D.C.

    """

    SA, CT, p = list(map(np.asanyarray, (SA, CT, p)))
    SA, CT, p = np.broadcast_arrays(SA, CT, p)
    SA, CT, p = list(map(ma.masked_invalid, (SA, CT, p)))

    p_min, idx = p.min(), p.argmin()

    sigma = gsw.rho(SA, CT, p_min) - 1000.0

    # Temperature and Salinity at the surface,
    T0, S0, Sig0 = CT[idx], SA[idx], sigma[idx]

    # NOTE: The temperature difference criterion for MLD
    Tdiff = T0 - 0.5  # 0.8 on the matlab original

    if criterion == "temperature":
        idx_mld = CT > Tdiff
    elif criterion == "pdvar":
        pdvar_diff = gsw.rho(S0, Tdiff, p_min) - 1000.0
        idx_mld = sigma <= pdvar_diff
    elif criterion == "density":
        sig_diff = Sig0 + 0.125
        idx_mld = sigma <= sig_diff
    else:
        raise NameError("Unknown criteria {}".format(criterion))

    MLD = ma.masked_all_like(p)
    MLD[idx_mld] = p[idx_mld]

    return MLD.max(axis=0), idx_mld


def pcaben(u, v):
    """
    Principal components of 2-d (e.g. current meter) data
    calculates ellipse parameters for currents.

    Parameters
    ----------
    u : array_like
        zonal wind velocity [m s :sup:`-1`]
    v : array_like
        meridional wind velocity [m s :sup:`-1`]

    Returns
    -------
    major axis (majrax)
    minor axis (minrax)
    major azimuth (majaz)
    minor azimuth (minaz)
    elipticity (elptcty)

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> from oceans.ocfis import pcaben, uv2spdir
    >>> u = np.r_[(0., 1., -2., -1., 1.), np.random.randn(10)]
    >>> v = np.r_[(3., 1., 0., -1., -1.), np.random.randn(10)]
    >>> (mjrax, mjaz, mirax, miaz, el), (x1, x2, y1, y2) = pcaben(u, v)
    >>> fig, ax = plt.subplots()
    >>> _ = ax.plot(x1, y1,'r-', x2, y2, 'r-')
    >>> ax.set_aspect('equal')
    >>> _ = ax.set_xlabel('U component')
    >>> _ = ax.set_ylabel('V component')
    >>> _ = ax.plot(u, v, 'bo')
    >>> _ = ax.axis([-3.2, 3.2, -3.2, 3.2])
    >>> mdir, mspd = uv2spdir(u.mean(), v.mean())
    >>> _ = ax.plot([0, u.mean()],[0, v.mean()], 'k-')
    >>> flatness = 1 - mirax / mjrax
    >>> if False:
    ...     print('Mean u = {}'.format(u.mean()))
    ...     print('Mean v = {}'.format(v.mean()))
    ...     print('Mean magnitude: {}'.format(mspd))
    ...     print('Direction: {}'.format(mdir))
    ...     print('Axis 1 Length: {}'.format(mjrax))
    ...     print('Axis 1 Azimuth: {}'.format(mjaz))
    ...     print('Axis 2 Length: {}'.format.format(mirax))
    ...     print('Axis 2 Azimuth: {}'.format(miaz))
    ...     print('elipticity is : {}'.format(el))
    ...     print('Negative (positive) means clockwise (anti-clockwise)')
    ...     print('Flatness: {}'.format(flatness))

    Notes:
    https://pubs.usgs.gov/of/2002/of02-217/m-files/pcaben.m

    """

    u, v = np.broadcast_arrays(u, v)

    C = np.cov(u, v)
    D, V = np.linalg.eig(C)

    x1 = np.r_[0.5 * np.sqrt(D[0]) * V[0, 0], -0.5 * np.sqrt(D[0]) * V[0, 0]]
    y1 = np.r_[0.5 * np.sqrt(D[0]) * V[1, 0], -0.5 * np.sqrt(D[0]) * V[1, 0]]
    x2 = np.r_[0.5 * np.sqrt(D[1]) * V[0, 1], -0.5 * np.sqrt(D[1]) * V[0, 1]]
    y2 = np.r_[0.5 * np.sqrt(D[1]) * V[1, 1], -0.5 * np.sqrt(D[1]) * V[1, 1]]

    # Length and direction.
    az, leng = np.c_[uv2spdir(x1[0], y1[0]), uv2spdir(x2[1], y2[1])]

    if leng[0] >= leng[1]:
        majrax, majaz = leng[0], az[0]
        minrax, minaz = leng[1], az[1]
    else:
        majrax, majaz = leng[1], az[1]
        minrax, minaz = leng[0], az[0]

    elptcty = minrax / majrax

    # Radius to diameter.
    majrax *= 2
    minrax *= 2

    return (majrax, majaz, minrax, minaz, elptcty), (x1, x2, y1, y2)


def spec_rot(u, v):
    """
    Compute the rotary spectra from u,v velocity components

    Parameters
    ----------
    u : array_like
        zonal wind velocity [m s :sup:`-1`]
    v : array_like
        meridional wind velocity [m s :sup:`-1`]

    Returns
    -------
    cw : array_like
         Clockwise spectrum [TODO]
    ccw : array_like
          Counter-clockwise spectrum [TODO]
    puv : array_like
          Cross spectra [TODO]
    quv : array_like
          Quadrature spectra [ TODO]

    Notes
    -----
    The spectral energy at some frequency can be decomposed into two circularly
    polarized constituents, one rotating clockwise and other anti-clockwise.

    Examples
    --------
    TODO: puv, quv, cw, ccw = spec_rot(u, v)

    References
    ----------
    J. Gonella Deep Sea Res., 833-846, 1972.

    """

    # Individual components Fourier series.
    fu, fv = list(map(np.fft.fft, (u, v)))

    # Auto-spectra of the scalar components.
    pu = fu * np.conj(fu)
    pv = fv * np.conj(fv)

    # Cross spectra.
    puv = fu.real * fv.real + fu.imag * fv.imag

    # Quadrature spectra.
    quv = -fu.real * fv.imag + fv.real * fu.imag

    # Rotatory components
    # TODO: Check the division, 4 (original code) or 8 (paper)?
    cw = (pu + pv - 2 * quv) / 4.0
    ccw = (pu + pv + 2 * quv) / 4.0
    N = len(u)
    F = np.arange(0, N) / N
    return puv, quv, cw, ccw, F


def lagcorr(x, y, M=None):
    """
    Compute lagged correlation between two series.
    Follow emery and Thomson book "summation" notation.

    Parameters
    ----------
    y : array
        time-series
    y : array
        time-series
    M : integer
        number of lags

    Returns
    -------
    Cxy : array
          normalized cross-correlation function

    Examples
    --------
    TODO: Implement Emery and Thomson example.

    """

    x, y = list(map(np.asanyarray, (x, y)))

    if not M:
        M = x.size

    N = x.size
    Cxy = np.zeros(M)
    x_bar, y_bar = x.mean(), y.mean()

    for k in range(0, M, 1):
        a = 0.0
        for i in range(N - k):
            a = a + (y[i] - y_bar) * (x[i + k] - x_bar)

        Cxy[k] = 1.0 / (N - k) * a

    return Cxy / (np.std(y) * np.std(x))


def complex_demodulation(series, f, fc, axis=-1):
    """
    Perform a Complex Demodulation
    It acts as a bandpass filter for `f`.

    series => Time-Series object with data and datetime
    f => inertial frequency in rad/sec
    fc => normalized cutoff [0.2]

    math : series.data * np.exp(2 * np.pi * 1j * (1 / T) * time_in_seconds)

    """
    import scipy.signal as signal

    # Time period ie freq = 1 / T
    T = 2 * np.pi / f
    # fc = fc * 1 / T # Filipe

    # De-mean data.
    d = series.data - series.data.mean(axis=axis)
    dfs = d * np.exp(2 * np.pi * 1j * (1 / T) * series.time_in_seconds)

    # Make a 5th order butter filter.
    # FIXME: Why 5th order? Try signal.buttord!?
    Wn = fc / series.Nyq

    [b, a] = signal.butter(5, Wn, btype="low")

    # FIXME: These are a factor of a thousand different from Matlab, why?
    cc = signal.filtfilt(b, a, dfs)  # FIXME: * 1e3
    amplitude = 2 * np.abs(cc)

    # TODO: Fix the outputs.
    # phase = np.arctan2(np.imag(cc), np.real(cc))

    filtered_series = amplitude * np.exp(
        -2 * np.pi * 1j * (1 / T) * series.time_in_seconds
    )
    new_series = filtered_series.real, series.time
    # Return cc, amplitude, phase, dfs, filtered_series
    return new_series


def binave(datain, r):
    """
    Averages vector data in bins of length r. The last bin may be the
    average of less than r elements. Useful for computing daily average time
    series (with r=24 for hourly data).

    Parameters
    ----------
    data : array_like
    r : int
        bin length

    Returns
    -------
    bindata : array_like
              bin-averaged vector

    Notes
    -----
    Original from MATLAB AIRSEA TOOLBOX.

    Examples
    --------
    >>> from oceans.ocfis import binave
    >>> data = [10., 11., 13., 2., 34., 21.5, 6.46, 6.27, 7.0867, 15., 123.,
    ...         3.2, 0.52, 18.2, 10., 11., 13., 2., 34., 21.5, 6.46, 6.27,
    ...         7.0867, 15., 123., 3.2, 0.52, 18.2, 10., 11., 13., 2., 34.,
    ...         21.5, 6.46, 6.27, 7.0867, 15., 123., 3.2, 0.52, 18.2, 10.,
    ...         11., 13., 2., 34., 21.5, 6.46, 6.27, 7.0867, 15., 123., 3.2,
    ...         0.52, 18.2]
    >>> binave(data, 24)
    array([16.564725 , 21.1523625, 22.4670875])

    References
    ----------
    03/08/1997: version 1.0
    09/19/1998: version 1.1 (vectorized by RP)
    08/05/1999: version 2.0

    """

    datain, rows = np.asarray(datain), np.asarray(r, dtype=np.int)

    if datain.ndim != 1:
        raise ValueError("Must be a 1D array.")

    if rows <= 0:
        raise ValueError("Bin size R must be a positive integer.")

    # Compute bin averaged series.
    lines = datain.size // r
    z = datain[0 : (lines * rows)].reshape(rows, lines, order="F")
    bindata = np.r_[np.mean(z, axis=0), np.mean(datain[(lines * r) :])]

    return bindata


def binavg(x, y, db):
    """
    Bins y(x) into db spacing.  The spacing is given in `x` units.
    y = np.random.random(20)
    x = np.arange(len(y))
    xb, yb = binavg(x, y, 2)
    plt.figure()
    plt.plot(x, y, 'k.-')
    plt.plot(xb, yb, 'r.-')

    """
    # Cut the corners.
    x_min, x_max = np.ceil(x.min()), np.floor(x.max())
    x = x.clip(x_min, x_max)

    # This is used to get the `inds`.
    xbin = np.arange(x_min, x_max, db)
    inds = np.digitize(x, xbin)

    # But this is the center of the bins.
    xbin = xbin - (db / 2.0)

    # FIXME there is an IndexError if I want to show this.
    # for n in range(x.size):
    #    print xbin[inds[n]-1], "<=", x[n], "<", xbin[inds[n]]

    ybin = np.array([y[inds == i].mean() for i in range(0, len(xbin))])
    # xbin = np.array([x[inds == i].mean() for i in range(0, len(xbin))])

    return xbin, ybin


def bin_dates(self, freq, tz=None):
    """
    Take a pandas time Series and return a new Series on the specified
    frequency.

    Examples
    --------
    >>> import numpy as np
    >>> from pandas import Series, date_range
    >>> n = 24*30
    >>> sig = np.random.rand(n) + 2 * np.cos(2 * np.pi * np.arange(n))
    >>> dates = date_range(start='1/1/2000', periods=n, freq='H')
    >>> series = Series(data=sig, index=dates)
    >>> new_series = bin_dates(series, freq='D', tz=None)

    """
    from pandas import date_range

    new_index = date_range(
        start=self.index[0], end=self.index[-1], freq=freq, tz=tz
    )
    new_series = self.groupby(new_index.asof).mean()
    # Averages at the center.
    secs = new_index.freq.delta.total_seconds()
    new_series.index = new_series.index.values + int(secs // 2)
    return new_series


def series_spline(self):
    """
    Fill NaNs using a spline interpolation.

    """
    from pandas import Series, isnull
    from scipy.interpolate import InterpolatedUnivariateSpline

    inds, values = np.arange(len(self)), self.values

    invalid = isnull(values)
    valid = -invalid

    firstIndex = valid.argmax()
    valid = valid[firstIndex:]
    invalid = invalid[firstIndex:]
    inds = inds[firstIndex:]

    result = values.copy()
    s = InterpolatedUnivariateSpline(inds[valid], values[firstIndex:][valid])
    result[firstIndex:][invalid] = s(inds[invalid])

    return Series(result, index=self.index, name=self.name)


def despike(self, n=3, recursive=False):
    """
    Replace spikes with np.NaN.
    Removing spikes that are >= n * std.
    default n = 3.

    """
    from pandas import Series

    result = self.values.copy()
    outliers = np.abs(self.values - np.nanmean(self.values)) >= n * np.nanstd(
        self.values
    )

    removed = np.count_nonzero(outliers)
    result[outliers] = np.NaN

    counter = 0
    if recursive:
        while outliers.any():
            result[outliers] = np.NaN
            base = np.abs(result - np.nanmean(result))
            outliers = base >= n * np.nanstd(result)
            counter += 1
            removed += np.count_nonzero(outliers)
    return Series(result, index=self.index, name=self.name)


def pol2cart(theta, radius, units="deg"):
    """
    Convert from polar to Cartesian coordinates

    Examples
    --------
    >>> pol2cart(0, 1, units='deg')
    (1.0, 0.0)

    """
    if units in ["deg", "degs"]:
        theta = theta * np.pi / 180.0
    x = radius * np.cos(theta)
    y = radius * np.sin(theta)
    return x, y


def cart2pol(x, y):
    """
    Convert from Cartesian to polar coordinates.

    Examples
    --------
    >>> x = [+0, -0.5]
    >>> y = [+1, +0.5]
    >>> cart2pol(x, y)
    (array([1.57079633, 2.35619449]), array([1.        , 0.70710678]))

    """
    radius = np.hypot(x, y)
    theta = np.arctan2(y, x)
    return theta, radius


def wrap_lon180(lon):
    lon = np.atleast_1d(lon).copy()
    angles = np.logical_or((lon < -180), (180 < lon))
    lon[angles] = wrap_lon360(lon[angles] + 180) - 180
    return lon


def wrap_lon360(lon):
    lon = np.atleast_1d(lon).copy()
    positive = lon > 0
    lon = lon % 360
    lon[np.logical_and(lon == 0, positive)] = 360
    return lon


def alphanum_key(s):
    key = re.split(r"(\d+)", s)
    key[1::2] = list(map(int, key[1::2]))
    return key


def get_profile(x, y, f, xi, yi, mode="nearest", order=3):
    """
    Interpolate regular data.

    Parameters
    ----------
    x : two dimensional np.ndarray
        an array for the :math:`x` coordinates

    y : two dimensional np.ndarray
        an array for the :math:`y` coordinates

    f : two dimensional np.ndarray
        an array with the value of the function to be interpolated
        at :math:`x,y` coordinates.

    xi : one dimension np.ndarray
        the :math:`x` coordinates of the point where we want
        the function to be interpolated.

    yi : one dimension np.ndarray
        the :math:`y` coordinates of the point where we want
        the function to be interpolated.

    order : int
        the order of the bivariate spline interpolation


    Returns
    -------
    fi : one dimension np.ndarray
        the value of the interpolating spline at :math:`xi,yi`


    Examples
    --------
    >>> import numpy as np
    >>> from oceans.ocfis import get_profile
    >>> x, y = np.meshgrid(range(360), range(91))
    >>> f = np.array(range(91 * 360)).reshape((91, 360))
    >>> Paris = [2.4, 48.9]
    >>> Rome = [12.5, 41.9]
    >>> Greenwich = [0, 51.5]
    >>> xi = Paris[0], Rome[0], Greenwich[0]
    >>> yi = Paris[1], Rome[1], Greenwich[1]
    >>> get_profile(x, y, f, xi, yi, order=3)
    array([17606, 15096, 18540])

    """
    from scipy.ndimage import map_coordinates

    x, y, f, xi, yi = list(map(np.asanyarray, (x, y, f, xi, yi)))
    conditions = np.array(
        [
            xi.min() < x.min(),
            xi.max() > x.max(),
            yi.min() < y.min(),
            yi.max() > y.max(),
        ]
    )

    if conditions.any():
        warnings.warn("Warning! Extrapolation!!")

    dx = x[0, 1] - x[0, 0]
    dy = y[1, 0] - y[0, 0]

    jvals = (xi - x[0, 0]) / dx
    ivals = (yi - y[0, 0]) / dy

    coords = np.array([ivals, jvals])

    return map_coordinates(f, coords, mode=mode, order=order)


def strip_mask(arr, fill_value=np.NaN):
    """
    Take a masked array and return its data(filled) + mask.

    """
    if ma.isMaskedArray(arr):
        mask = np.ma.getmaskarray(arr)
        arr = np.ma.filled(arr, fill_value)
        return mask, arr
    else:
        return arr


def shiftdim(x, n=None):
    """
    Matlab-like shiftdim in python.

    Examples
    --------
    >>> import oceans.ocfis as ff
    >>> a = np.random.rand(1,1,3,1,2)
    >>> print("a shape and dimension: %s, %s" % (a.shape, a.ndim))
    a shape and dimension: (1, 1, 3, 1, 2), 5
    >>> # print(range(a.ndim))
    >>> # print(np.roll(range(a.ndim), -2))
    >>> # print(a.transpose(np.roll(range(a.ndim), -2)))
    >>> b = ff.shiftdim(a)
    >>> print("b shape and dimension: %s, %s" % (b.shape, b.ndim))
    b shape and dimension: (3, 1, 2), 3
    >>> c = ff.shiftdim(b, -2)
    >>> c.shape == a.shape
    True

    """

    def no_leading_ones(shape):
        shape = np.atleast_1d(shape)
        if shape[0] == 1:
            shape = shape[1:]
            return no_leading_ones(shape)
        else:
            return shape

    if n is None:
        # returns the array B with the same number of
        # elements as X but with any leading singleton
        # dimensions removed.
        return x.reshape(no_leading_ones(x.shape))
    elif n >= 0:
        # When n is positive, shiftdim shifts the dimensions
        # to the left and wraps the n leading dimensions to the end.
        return x.transpose(np.roll(list(range(x.ndim)), -n))
    else:
        # When n is negative, shiftdim shifts the dimensions
        # to the right and pads with singletons.
        return x.reshape((1,) * -n + x.shape)



import os

from setuptools import find_packages, setup

import versioneer


rootpath = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    return open(os.path.join(rootpath, *parts), "r").read()


email = "ocefpaf@gmail.com"
maintainer = "Filipe Fernandes"
authors = [u"André Palóczy", "Arnaldo Russo", "Filipe Fernandes"]

# Dependencies.
hard = ["gsw", "matplotlib", "numpy", "seawater"]
soft = {"full": ["cartopy", "iris", "netcdf4", "pandas", "scipy"]}

setup(
    name="oceans",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    packages=find_packages(),
    package_data={"oceans": ["colormaps/cmap_data/*.dat"]},
    license="BSD-3-Clause",
    long_description=f'{read("README.md")}',
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Topic :: Education",
        "Topic :: Scientific/Engineering",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    description="Misc functions for oceanographic data analysis",
    author=authors,
    author_email=email,
    maintainer="Filipe Fernandes",
    maintainer_email=email,
    url="https://pypi.python.org/pypi/oceans/",
    platforms="any",
    keywords=["oceanography", "data analysis"],
    extras_require=soft,
    install_requires=hard,
    python_requires='>=3.6',
    tests_require=["pytest"],
)



import os

from colorsys import hsv_to_rgb
from glob import glob

import matplotlib.pyplot as plt
import numpy as np

from matplotlib import colors


cmap_path = os.path.join(os.path.dirname(__file__), "cmap_data")


class Bunch(dict):
    def __init__(self, **kw):
        dict.__init__(self, kw)
        self.__dict__ = self


def get_color(color):
    """
    https://stackoverflow.com/questions/10254207/color-and-line-writing-using-matplotlib

    """
    for hue in range(color):
        hue = 1.0 * hue / color
        col = [int(x) for x in hsv_to_rgb(hue, 1.0, 230)]
        yield "#{0:02x}{1:02x}{2:02x}".format(*col)


def cmat2cmpl(rgb, reverse=False):
    """
    Convert RGB matplotlib colormap.

    """
    rgb = np.asanyarray(rgb)
    if reverse:
        rgb = np.flipud(rgb)
    return colors.ListedColormap(rgb)


def phasemap_cm(m=256):
    """
    Colormap periodic/circular data (phase).

    """

    theta = 2 * np.pi * np.arange(0, m) / m
    circ = np.exp(1j * theta)

    # Vertices of colour triangle.
    vred, vgreen, vblue = -2, 1 - np.sqrt(3) * 1j, 1 + np.sqrt(3) * 1j

    vredc = vred - circ
    vgreenc = vgreen - circ
    vbluec = vblue - circ

    red = np.abs(np.imag(vgreenc * np.conj(vbluec)))
    green = np.abs(np.imag(vbluec * np.conj(vredc)))
    blue = np.abs(np.imag(vredc * np.conj(vgreenc)))

    return (
        1.5
        * np.c_[red, green, blue]
        / np.abs(np.imag((vred - vgreen) * np.conj(vred - vblue)))
    )


def zebra_cm(a=4, m=0.5, n=256):
    """
    Zebra palette colormap with NBANDS broad bands and NENTRIES rows in
    the color map.

    The default is 4 broad bands

    cmap = zebra(nbands, nentries)

    References
    ----------
    Hooker, S. B. et al, Detecting Dipole Ring Separatrices with Zebra
    Palettes, IEEE Transactions on Geosciences and Remote Sensing, vol. 33,
    1306-1312, 1995

    Notes
    -----
    Saturation and value go from m to 1 don't use m = 0.
    a = 4 -> there are this many large bands in the palette.

    """
    from scipy.signal import sawtooth

    x = np.arange(0, n)
    hue = np.exp(-3.0 * x / n)
    sat = m + (1.0 - m) * (0.5 * (1.0 + sawtooth(2.0 * np.pi * x / (n / a))))
    val = m + (1.0 - m) * 0.5 * (1.0 + np.cos(2.0 * np.pi * x / (n / a / 2.0)))
    return np.array([hsv_to_rgb(h, s, v) for h, s, v in zip(hue, sat, val)])


def ctopo_pos_neg_cm(m=256):
    """
    Colormap for positive/negative data with gray scale only
    original from cushman-roisin book cd-rom.

    """
    dx = 1.0 / (m - 1)
    values = np.arange(0.0, 1.0, dx)
    return np.c_[values, values, values]


def avhrr_cm(m=256):
    """
    AHVRR colormap used by NOAA Coastwatch.

    """

    x = np.arange(0.0, m) / (m - 1)

    xr = [0.0, 0.2, 0.4, 0.5, 0.6, 0.8, 1.0]
    rr = [0.5, 1.0, 1.0, 0.5, 0.5, 0.0, 0.5]

    xg = [0.0, 0.4, 0.6, 1.0]
    gg = [0.0, 1.0, 1.0, 0.0]

    xb = [0.0, 0.2, 0.4, 0.5, 0.6, 0.8, 1.0]
    bb = [0.0, 0.0, 0.5, 0.5, 1.0, 1.0, 0.5]

    r = np.interp(x, xr, rr)
    g = np.interp(x, xg, gg)
    b = np.interp(x, xb, bb)

    return np.flipud(np.c_[r, g, b])


def load_cmap(fname):
    return np.loadtxt(fname, delimiter=",") / 255


# Functions colormaps.
arrays = {
    "zebra": zebra_cm(),
    "avhrr": avhrr_cm(),
    "phasemap": phasemap_cm(),
    "ctopo_pos_neg": ctopo_pos_neg_cm(),
}

# Data colormaps.
for fname in glob("%s/*.dat" % cmap_path):
    cmap = os.path.basename(fname).split(".")[0]
    data = load_cmap(fname)
    arrays.update({cmap: data})

cm = Bunch()
for key, value in arrays.items():
    cm.update({key: cmat2cmpl(value)})
    cm.update({"%s_r" % key: cmat2cmpl(value, reverse=True)})


def demo():
    data = np.outer(np.arange(0, 1, 0.01), np.ones(10))
    fig = plt.figure(figsize=(10, 5))
    fig.subplots_adjust(top=0.8, bottom=0.05, left=0.01, right=0.99)
    cmaps = sorted((m for m in cm.keys() if not m.endswith("_r")))
    length = len(cmaps)
    for k, cmap in enumerate(cmaps):
        plt.subplot(1, length + 1, k + 1)
        plt.axis("off")
        plt.imshow(data, aspect="auto", cmap=cm.get(cmap), origin="lower")
        plt.title(cmap, rotation=90, fontsize=10)
